/**
 * @file blue.c
 *
 * This is the top-level file that provides the initial entry point
 * to the QNX Bluetooth stack.
 *
 * Revision History:
 *   2003 Feb 4   David Moore      Initial version.
 *   2003 Mar 20  David Moore      Code cleanups and documentation
 */

#include <stdio.h>
#include "blue_driver.h"

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>

/** The master data structure for the Bluetooth driver */
blue_t BlueCtrl;

/**
 * Frees a single Bluetooth interface (device).
 */
int blue_iface_free(blue_iface_t * iface)
{
    iface_resmgr_detach(iface);
    hci_send_command(iface, HCI_OGF_LINK_CONTROL,
            HCI_EXIT_PER_INQUIRY, NULL, 0);
    hci_free(iface);
    return blue_usb_free(iface);
}

/**
 * Frees the entire Bluetooth driver and all attached interfaces.
 */
int blue_free(blue_t * blue)
{
    blue_iface_t * iface, * niface;

    /* Stop the L2CAP API */
    l2cap_resmgr_detach(blue);

    /* Free each interface (device) */
    for (iface = TAILQ_FIRST(&blue->interfaces); iface; iface = niface) {
        niface = TAILQ_NEXT(iface, dlink);
        blue_iface_free(iface);
    }

    return EOK;
}

/**
 * Puts the driver in daemon mode and sets up a dispatcher for the
 * resource manager.
 */
static int blue_dispatch_init()
{
    resmgr_attr_t   attr;

    /* Become a daemon by running in the background */
    if (procmgr_daemon(EXIT_SUCCESS, PROCMGR_DAEMON_NOCLOSE |
                PROCMGR_DAEMON_NODEVNULL) == -1) {
        SLOG(_SLOG_ERROR, "bluetooth: Failed to run as daemon");
        return -1;
    }

    /* Create a dispatcher.  This code is basically taken verbatim from the
     * QNX resource manager documentation. */
    if (!(BlueCtrl.dispatch = dispatch_create())) {
        SLOG(_SLOG_ERROR, "bluetooth: Unable to allocate dispatch context");
        return -1;
    }

    memset(&BlueCtrl.pool_attr, 0, sizeof(BlueCtrl.pool_attr));
    BlueCtrl.pool_attr.attr = malloc(sizeof(*BlueCtrl.pool_attr.attr));
    if (!BlueCtrl.pool_attr.attr) {
        SLOG(_SLOG_ERROR, "bluetooth: Unable to allocate pool attributes");
        return -1;
    }

    pthread_attr_init(BlueCtrl.pool_attr.attr);
    pthread_attr_setstacksize(BlueCtrl.pool_attr.attr, 8192);

    BlueCtrl.pool_attr.handle          = BlueCtrl.dispatch;
    BlueCtrl.pool_attr.context_alloc   = resmgr_context_alloc;
    BlueCtrl.pool_attr.block_func      = resmgr_block;
    BlueCtrl.pool_attr.handler_func    = (void *)dispatch_handler;
    BlueCtrl.pool_attr.context_free    = resmgr_context_free;
    BlueCtrl.pool_attr.lo_water        = 1;
    BlueCtrl.pool_attr.hi_water        = 2;
    BlueCtrl.pool_attr.increment       = 1;
    BlueCtrl.pool_attr.maximum         = 2;

    BlueCtrl.thread_pool = thread_pool_create(&BlueCtrl.pool_attr, 0);
    if (!BlueCtrl.thread_pool) {
        SLOG(_SLOG_ERROR, "bluetooth: Unable to initialize thread pool");
        return -1;
    }

    memset(&attr, 0, sizeof(attr));
    attr.nparts_max = 10;
    attr.msg_max_size = 0;

    if (resmgr_attach(BlueCtrl.dispatch, &attr, NULL, 0, 0, NULL, NULL, NULL) == -1) {
        SLOG(_SLOG_ERROR, "bluetooth: Unable to attach to resmgr");
        return -1;
    }

    if (thread_pool_start(BlueCtrl.thread_pool)) {
        SLOG(_SLOG_ERROR, "bluetooth: Unable to start thread pool");
        return -1;
    }

    return 0;
}

/**
 * Handles the TERM signal by freeing the Bluetooth driver and quitting.
 */
void blue_sigterm_handler(int signo)
{
    SLOG(_SLOG_INFO, "bluetooth: Quitting...");
    blue_free(&BlueCtrl);
    exit(EXIT_SUCCESS);
}

/**
 * Sets a signal handler for TERM, and block all others.
 */
void blue_set_signal_handler()
{
    sigset_t    set;

    sigfillset(&set);
    sigdelset(&set, SIGTERM);
    sigprocmask(SIG_BLOCK, &set, NULL);

    sigemptyset(&set);
    sigaddset(&set, SIGTERM);
    sigprocmask(SIG_UNBLOCK, &set, NULL);

    signal(SIGTERM, blue_sigterm_handler);
}

/**
 * Initializes the Bluetooth driver.
 */
int blue_start_driver (int argc, char **argv) {
    sigset_t            signals;

    /* Block most signals */
    sigfillset(&signals);
    pthread_sigmask(SIG_BLOCK, &signals, NULL);

    /* Initializes our list of device interfaces */
    TAILQ_INIT(&BlueCtrl.interfaces);

    /* Start a resource manager/dispatcher for the L2CAP socket-based API */
    if (blue_dispatch_init()) {
        return -1;
    }
    if (l2cap_resmgr_attach(&BlueCtrl) != EOK) {
        return -1;
    }

    /* Attach to the USB stack */
    if (blue_usb_start_driver(&BlueCtrl, argc, argv) != EOK) {
        return -1;
    }

    SLOG(_SLOG_INFO, "bluetooth: QNX Bluetooth class driver v%s", VERSION);
    return EOK;
}

/**
 * Entry point for a thread that processes the command queue for a
 * single Bluetooth interface.  HCI commands issued by higher layers
 * are queued in memory.  This thread feeds each command to the
 * hardware one-by-one through the USB command endpoint, observing
 * the hardware's maximum number of concurrent commands.
 */
static void * cmd_queue_thread(void * arg)
{
    blue_iface_t  *iface = (blue_iface_t *) arg;
    databuf_t     *cmd;
    unsigned char *buf;

    while (1) {
        pthread_mutex_lock(&iface->mutex);
        /* Get the first available command */
        while ((cmd = TAILQ_FIRST(&iface->cmd_q))) {
            /* If the device has already received its maximum number of
             * commands, wait for them to finish */
            while (iface->num_hci_cmd_packets <= 0) {
                pthread_mutex_unlock(&iface->mutex);

                pthread_sleepon_lock();
                pthread_sleepon_wait(&iface->num_hci_cmd_packets);
                pthread_sleepon_unlock();

                pthread_mutex_lock(&iface->mutex);
            }
            
            /* Remove the command from the queue */
            TAILQ_REMOVE(&iface->cmd_q, cmd, dlink);
            pthread_mutex_unlock(&iface->mutex);

            buf = usbd_alloc(cmd->len);
            if (buf) {
                memcpy(buf, cmd->data, cmd->len);

                pthread_mutex_lock(&iface->mutex);
                iface->num_hci_cmd_packets--;
                pthread_mutex_unlock(&iface->mutex);
                
                /* Send the command to the hardware.  This function
                 * blocks until the command is fully received. */
                blue_cmd_io(iface, iface->ep_cntl,
                        URB_DIR_OUT, buf, cmd->len);
                usbd_free(buf);
            }
            else {
                SLOG(_SLOG_CRITICAL, "bluetooth: Out of memory in command queue");
            }
            free(cmd);
            pthread_mutex_lock(&iface->mutex);
        }
        pthread_mutex_unlock(&iface->mutex);

        /* Wait for any new commands */
        pthread_sleepon_lock();
        pthread_sleepon_wait(&iface->cmd_q);
        pthread_sleepon_unlock();
    }
}

/**
 * Entry point for a thread that processes the data packet queue for a
 * single Bluetooth interface.  ACL packets issued by higher layers
 * are queued in memory.  This thread feeds each packet to the
 * hardware one-by-one through the USB bulk endpoint, observing
 * the hardware's maximum number of concurrent packets.
 */
static void * acl_queue_thread(void * arg)
{
    blue_iface_t  *iface = (blue_iface_t *) arg;
    hci_conn_t    *conn;
    databuf_t     *acl;
    unsigned char *buf;
    int            num_packets;

    while (1) {
        /* Determine the next connection handle that needs a packet sent.
         * hci_sched_next_acl() tries to schedule connections to avoid
         * filling the queue with stuck packets.
         * TODO: some way of getting this unstuck when the buffer is full
         *       for too long. */
        while ((conn = hci_sched_next_acl(iface, &num_packets))) {
            pthread_mutex_lock(&conn->acl_mutex);

            /* Send the allowed number of packets for the connection handle */
            while (num_packets && (acl = TAILQ_FIRST(&conn->acl_q))) {
                hci_acldata_t * hdr = (hci_acldata_t *) acl->data;

                /* Remove the packet from the queue */
                TAILQ_REMOVE(&conn->acl_q, acl, dlink);
                pthread_mutex_unlock(&conn->acl_mutex);

                buf = usbd_alloc(acl->len);
                if (!buf) {
                    free(acl);
                    continue;
                }

                /* Prepend the connection handle to the packet */
                hdr->handle = ENDIAN_LE16(conn->handle);
                memcpy(buf, acl->data, acl->len);
                SLOG(_SLOG_DEBUG1, "bluetooth: Sending ACL packet of length %d",
                        acl->len);

                /* Temporarily suspend the bulk endpoint listener (QNX does not
                 * seem to allow us to send and receive on the endpoint
                 * simultaneously) */
                blue_abort_acl_listen(iface);
                /* Submit the packet to the hardware */
                blue_bulk_io(iface, iface->ep_bulk_out,
                        URB_DIR_OUT | URB_SHORT_XFER_OK, buf, acl->len);
                /* Restart the bulk listener */
                blue_start_acl_listen(iface);

                usbd_free(buf);
                free(acl);

                /* Decrement the number of available packet slots on the device */
                pthread_mutex_lock(&iface->mutex);
                iface->num_hci_acl_packets--;
                pthread_mutex_unlock(&iface->mutex);
                
                pthread_mutex_lock(&conn->acl_mutex);
                num_packets--;
                conn->acl_pending++;
            }
            pthread_mutex_unlock(&conn->acl_mutex);
        }
        
        /* Wait for a new packet */
        pthread_sleepon_lock();
        pthread_sleepon_wait(&iface->num_hci_acl_packets);
        pthread_sleepon_unlock();
    }
}

/**
 * The initial entry point for the QNX Bluetooth stack.  This starts the
 * driver and monitors for device insertions.
 */
int main(int argc, char ** argv)
{
    /* Switches to daemon mode, attaches to the resource manager, attaches
     * to the USB stack, establishes callbacks for insertion/removal, and
     * creates a new socket type. */
    if (blue_start_driver(argc, argv)) {
        SLOG(_SLOG_ERROR, "bluetooth: Failed to start driver, aborting...");
        exit(EXIT_FAILURE);
    }

    /* Watch for termination signals */
    blue_set_signal_handler();

    /* This loop is executed once for each new device inserted in order to
     * perform proper initialization of the hardware. */
    while (1) {
        blue_iface_t * iface;

        /* Wait for a new device */
        pthread_sleepon_lock();
        if (pthread_sleepon_wait(&BlueCtrl.interfaces)) {
            break;
        }
        pthread_sleepon_unlock();
        
        /* Add the device to the list of interfaces */
        iofunc_attr_lock(&BlueCtrl.attr);
        iface = TAILQ_FIRST(&BlueCtrl.interfaces);
        iofunc_attr_unlock(&BlueCtrl.attr);
        
        if (iface && !(iface->status & BLUE_PRESENT)) {
            cmd_host_buffer_size_t  cmdp1;
            cmd_per_inquiry_t           cmdp2;
            unsigned char           scan;

            if (blue_load_firmware(iface))
                continue;

            iface->status = BLUE_PRESENT;

            /* Initialize callbacks for the USB endpoints and start
             * threads to monitor each endpoint. */
            iface->hci_event = hci_event;
            iface->hci_aclrecv = hci_aclrecv;
            blue_start_ep_listeners(iface);
            pthread_create(NULL, NULL, acl_queue_thread, iface);
            pthread_create(NULL, NULL, cmd_queue_thread, iface);

            /* Inform the device of the buffer size on the host for
             * ACL and SCO packets. */
            cmdp1.max_acl_len     = ENDIAN_LE16(MAX_RECV_ACL_LEN - 4);
            cmdp1.max_sco_len     = MAX_RECV_SCO_LEN - 3;
            cmdp1.max_acl_packets = ENDIAN_LE16(8);
            cmdp1.max_sco_packets = ENDIAN_LE16(8);
            hci_send_command(iface, HCI_OGF_HC_AND_BASEBAND,
                    HCI_HOST_BUFFER_SIZE, &cmdp1, sizeof(cmdp1));

            /* Read the device's buffer size */
            hci_send_command(iface, HCI_OGF_INFO,
                    HCI_READ_BUFFER_SIZE, NULL, 0);

            /* Configure the device to respond to inquiry and page
             * requests. */
            scan = SCAN_ENABLE_INQUIRY_AND_PAGE;
            hci_send_command(iface, HCI_OGF_HC_AND_BASEBAND,
                    HCI_WRITE_SCAN_ENABLE, &scan, sizeof(scan));

            /* Read the local Bluetooth Device Address */
            hci_send_command(iface, HCI_OGF_INFO,
                    HCI_READ_BDADDR, NULL, 0);

            /* Configure the device to perform periodic inquiries of
             * nearby devices. */
            cmdp2.max_period = ENDIAN_LE16(20);
            cmdp2.min_period = ENDIAN_LE16(15);
            cmdp2.lap = ENDIAN_LE16(LAP_GENERAL_INQUIRY);
            cmdp2.length = 5;
            cmdp2.num_responses = 0;
            hci_send_command(iface, HCI_OGF_LINK_CONTROL,
                    HCI_PER_INQUIRY, &cmdp2, sizeof(cmdp2));

            /* Create a configuration interface at /dev/blue/hciN */
            iface_resmgr_attach(iface);
        }
    }

    SLOG(_SLOG_ERROR, "bluetooth: Aborting");
    
    exit(EXIT_SUCCESS);
}
