#ifndef _INCLUDE_BLUE_DRIVER_H
#define _INCLUDE_BLUE_DRIVER_H

#include <share.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <gulliver.h>
#include <sys/usbdi.h>

#include <sys/iofunc.h>
#include <sys/resmgr.h>
#include <sys/procmgr.h>
#include <sys/dcmd_chr.h>
#include <sys/dispatch.h>
#include <sys/neutrino.h>
#include <sys/rsrcdbmgr.h>
#include <sys/slog.h>
#include <sys/slogcodes.h>

#include "queue.h"

#include <bluetooth/hci.h>
#include <bluetooth/l2cap.h>

typedef struct _l2cap_cl_pkt {
    unsigned short          len;
    unsigned short          chid;
    unsigned short          psm;
    unsigned char           data[0];
} __attribute__ ((packed)) l2cap_cl_pkt_t;

typedef struct _databuf {
    TAILQ_ENTRY(_databuf)   dlink;
    int                     len;
    unsigned char           data[0];
} databuf_t;

typedef struct _inq_ent {
    TAILQ_ENTRY(_inq_ent)   dlink;
    inq_res_t               i;
} inq_entry_t;

struct ocb_l2;

typedef struct _blue {
    iofunc_attr_t            attr;
    int                      resmgr_id;
    resmgr_connect_funcs_t   resmgr_connect_funcs;
    resmgr_io_funcs_t        resmgr_io_funcs;
    iofunc_mount_t           io_mount;
    struct termios           tios;
    TAILQ_HEAD(,ocb_l2)      sockets;
    TAILQ_HEAD(,_blue_iface) interfaces;
    iofunc_notify_t          notify[3];
    _uint32                         cflags;
    thread_pool_t                   *thread_pool;
    dispatch_t                      *dispatch;
    thread_pool_attr_t              pool_attr;
    int                             verbose;
} blue_t;

typedef struct _reply_info {
    int     rcvid;
    int     offset;
} reply_info_t;

struct _blue_iface;

typedef int (*hci_status_callback_t)(struct _blue_iface *, reply_info_t, unsigned char);
typedef int (*hci_event_callback_t)(struct _blue_iface *, reply_info_t, unsigned char *);

typedef struct _cbk_ent {
    unsigned short          opcode;
    unsigned char           event;
    reply_info_t            reply_info;
    hci_event_callback_t    event_callback;
    hci_status_callback_t   status_callback;
    TAILQ_ENTRY(_cbk_ent)   dlink;
} callback_entry_t;

typedef struct _blue_iface {
    iofunc_attr_t            attr;
    int                      resmgr_id;
    resmgr_connect_funcs_t   resmgr_connect_funcs;
    resmgr_io_funcs_t        resmgr_io_funcs;
    struct termios           tios;
    iofunc_mount_t           io_mount;
    int                      devnum;
    blue_t                 *blue;
    TAILQ_ENTRY(_blue_iface)    dlink;
    pthread_mutex_t         mutex;
    bdaddr_t                bdaddr;
    usbd_device_instance_t  instance;
    struct usbd_device     *device;
    _uint32                 status;
    struct usbd_pipe       *ep_cntl;
    struct usbd_pipe       *ep_int;
    struct usbd_pipe       *ep_bulk_in;
    struct usbd_pipe       *ep_bulk_out;
    uint32_t                ep_int_size;
    uint32_t                ep_bulk_size;
    struct usbd_urb        *urb_cmd;
    struct usbd_urb        *urb_int;
    struct usbd_urb        *urb_bulk_in;
    struct usbd_urb        *urb_bulk_out;
    _uint8                 *ep_int_buf;
    _uint8                 *ep_bulk_buf;
    _uint8                 *evt_buf;
    _uint8                 *acl_buf;
    int                     evt_buf_pos;
    int                     acl_buf_pos;
    TAILQ_HEAD(,_databuf)   cmd_q;
    int                     num_hci_cmd_packets;
    int                     num_hci_acl_packets;
    int                     max_acl_data_len;
    TAILQ_HEAD(,_hci_conn)  connections;
    TAILQ_HEAD(,_inq_ent)   inquiry_cache;
    TAILQ_HEAD(,_cbk_ent)   callbacks;
    int (*hci_event)  (struct _blue_iface *, unsigned char *, _uint32);
    int (*hci_aclrecv)(struct _blue_iface *, unsigned char *, _uint32);
} blue_iface_t;

struct ocb_l2 {
    iofunc_ocb_t            hdr;
    blue_t                 *blue;
    unsigned int            pkt_type;
    unsigned int            state;
    bdaddr_t                remote_bdaddr;
    unsigned short          remote_psm;
    bdaddr_t                local_bdaddr;
    unsigned short          local_psm;
    int                     is_bound;
    TAILQ_ENTRY(ocb_l2)     dlink;
    TAILQ_HEAD(,_l2cap_pkt) recv_pkts;
    int                     saved_rcvid;
    int                     saved_addrlen;
    int                     saved_nbytes;
};

typedef struct _l2cap_pkt {
    int                         len;
    struct sockaddr_l2          addr;
    TAILQ_ENTRY(_l2cap_pkt)     dlink;
    unsigned char               data[0];
} l2cap_pkt_t;

typedef struct _hci_conn {
    TAILQ_ENTRY(_hci_conn)  dlink;
    blue_iface_t           *iface;
    unsigned short          handle;
    bdaddr_t                bdaddr;
    pthread_mutex_t         acl_mutex;
    TAILQ_HEAD(,_databuf)   acl_q;
    unsigned int            acl_pending;
    unsigned int            recv_buf_pos;
    unsigned char           recv_buf[L2CAP_MTU + 4];
} hci_conn_t;


#define SK_UNCONNECTED    1
#define SK_CONNECTED      2

#define MAX_BLUE_DEVICES    16

#define BLUE_PRESENT        0x1
#define BLUE_UNBLOCK        0x2
#define BLUE_RECV_ABORTED   0x4
#define BLUE_READY          0x8

#define BLUE_CONTROL_EP     0x1
#define BLUE_INTIN_EP       0x2
#define BLUE_BULKIN_EP      0x4
#define BLUE_BULKOUT_EP     0x8

#define MAX_RECV_ACL_LEN    1024
#define MAX_RECV_SCO_LEN    256
#define MAX_RECV_EVT_LEN    260

extern blue_t BlueCtrl;

extern int blue_load_firmware(blue_iface_t * iface);
extern int blue_bulk_io(blue_iface_t * iface, struct usbd_pipe * pipe, _uint32 flags,
        _uint8 * buffer, _uint32 length);
extern int blue_cmd_io(blue_iface_t * iface, struct usbd_pipe * pipe, _uint32 flags,
        _uint8 * buffer, _uint32 length);
extern int blue_start_acl_listen(blue_iface_t * blue);
extern int blue_abort_acl_listen(blue_iface_t * blue);
extern int blue_intr_in_sync(blue_iface_t * iface);
extern int blue_start_ep_listeners(blue_iface_t * iface);
extern int blue_iface_free(blue_iface_t * iface);
extern int blue_free(blue_t * blue);
extern int blue_usb_free(blue_iface_t * iface);
extern int blue_usb_start_driver (blue_t * blue, int argc, char **argv);
extern int blue_resmgr_attach(blue_t * blue);

extern int hci_aclrecv(blue_iface_t * iface, unsigned char * buf, _uint32 len);
extern int hci_event  (blue_iface_t * iface, unsigned char * buf, _uint32 len);
extern int hci_send_command(blue_iface_t * iface, int ogf, int ocf,
        void * param, int param_len);
extern int hci_send_acl(hci_conn_t * conn, unsigned int flags,
        unsigned char * data, unsigned int len);
extern hci_conn_t * hci_sched_next_acl(blue_iface_t * iface, unsigned * num_packets);
extern hci_conn_t * hci_connect(blue_iface_t * iface, bdaddr_t bdaddr);
extern void hci_free(blue_iface_t * iface);
extern blue_iface_t * hci_get_default_iface(blue_t * blue);
extern blue_iface_t * hci_get_iface_by_bdaddr(blue_t * blue, bdaddr_t bdaddr);
extern inq_res_t * hci_find_inquiry_result(blue_iface_t * iface, bdaddr_t bdaddr);
extern int hci_queue_callback(blue_iface_t * iface, int ogf, int ocf, int event,
        reply_info_t reply_info, hci_status_callback_t status_callback,
        hci_event_callback_t event_callback);

extern int iface_resmgr_attach(blue_iface_t * blue);
extern void iface_resmgr_detach(blue_iface_t * blue);

extern int l2cap_resmgr_attach(blue_t * blue);
extern void l2cap_resmgr_detach(blue_t * blue);
extern int l2cap_recv_packet(hci_conn_t * conn,
        unsigned char * buf, unsigned int len);

#ifndef _SLOGC_USB_GEN
#define _SLOGC_USB_GEN	__C(_SLOGC_USB, 0)
#endif

#define SLOG(severity, fmt, ...) do { \
    if (severity > BlueCtrl.verbose + 5) \
        break; \
    slogf(_SLOGC_USB_GEN, severity, fmt, ## __VA_ARGS__); \
} while (0)

#endif
