/**
 * @file blue_usb.c
 *
 * USB-specific code for initializing a USB Bluetooth device.
 *
 * Revision History:
 *   2003 Feb 4   David Moore      Initial version.
 */

#include <stdio.h>
#include "blue_driver.h"


int blue_usb_free(blue_iface_t * iface)
{
    int status;
    
    iface->status &= ~BLUE_PRESENT;

    if (iface->ep_bulk_out) {
        usbd_abort_pipe(iface->ep_bulk_out);
    }

    if (iface->ep_bulk_in) {
        usbd_abort_pipe(iface->ep_bulk_in);
    }

    if (iface->ep_int) {
        usbd_abort_pipe(iface->ep_int);
    }

    /* Give time for the abort to proceed */
    usleep(50);
    
    pthread_sleepon_lock();
    pthread_sleepon_signal(&iface->urb_cmd);
    pthread_sleepon_signal(&iface->urb_int);
    pthread_sleepon_signal(&iface->urb_bulk_in);
    pthread_sleepon_signal(&iface->urb_bulk_out);
    pthread_sleepon_unlock();

    usbd_free(iface->ep_bulk_buf);
    usbd_free(iface->ep_int_buf);
    usbd_free_urb(iface->urb_cmd);
    usbd_free_urb(iface->urb_int);
    usbd_free_urb(iface->urb_bulk_in);
    usbd_free_urb(iface->urb_bulk_out);

    iofunc_attr_lock(&iface->blue->attr);
    TAILQ_REMOVE(&iface->blue->interfaces, iface, dlink);
    iofunc_attr_unlock(&iface->blue->attr);

    if ((status = usbd_detach(iface->device)) != EOK) {
        SLOG(_SLOG_NOTICE, "bluetooth: USB detach failed: %s", strerror(status));
    }

    return status;
}

int blue_parse_descriptors(blue_iface_t * iface)
{
    _uint32                 eix, scan, found;
    usbd_descriptors_t      *desc;
    struct usbd_desc_node   *ifc, *ept;

    scan  = BLUE_CONTROL_EP | BLUE_INTIN_EP | BLUE_BULKIN_EP | BLUE_BULKOUT_EP;
    found = 0;

    if (!usbd_interface_descriptor(iface->device, iface->instance.config, iface->instance.iface,
                iface->instance.alternate, &ifc))
        return ENODEV;

    for (eix = 0; ; eix++) {
        desc = usbd_parse_descriptors(iface->device, ifc, USB_DESC_ENDPOINT, eix, &ept);
        if (!desc)
            break;
        switch (desc->endpoint.bmAttributes) {
            case USB_ATTRIB_CONTROL:
                if (usbd_open_pipe(iface->device, desc, &iface->ep_cntl) == EOK)
                    found |= BLUE_CONTROL_EP;
                break;
            case USB_ATTRIB_BULK:
                switch (desc->endpoint.bEndpointAddress & USB_ENDPOINT_IN) {
                    case USB_ENDPOINT_OUT:
                        if (usbd_open_pipe(iface->device, desc, &iface->ep_bulk_out) == EOK) {
                            found |= BLUE_BULKOUT_EP;
                        }
                        break;
                    case USB_ENDPOINT_IN:
                        if (usbd_open_pipe(iface->device, desc, &iface->ep_bulk_in) == EOK) {
                            iface->ep_bulk_size = desc->endpoint.wMaxPacketSize;
                            found |= BLUE_BULKIN_EP;
                        }
                        break;
                }
                break;
            case USB_ATTRIB_INTERRUPT:
                switch (desc->endpoint.bEndpointAddress & USB_ENDPOINT_IN) {
                    case USB_ENDPOINT_IN:
                        if (usbd_open_pipe(iface->device, desc, &iface->ep_int) == EOK) {
                            iface->ep_int_size = desc->endpoint.wMaxPacketSize;
                            found |= BLUE_INTIN_EP;
                        }
                        break;
                }
                break;
            default:
                SLOG(_SLOG_NOTICE, "bluetooth: Unknown EP %x", desc->endpoint.bmAttributes);
        }
    }
    if (found != scan) {
        SLOG(_SLOG_ERROR, "bluetooth: Failed to find required endpoints");
        return ENODEV;
    }

    return EOK;
}

// this is the function called when the USB device is inserted
void blue_insert(struct usbd_connection * conn, usbd_device_instance_t *instance)
{
    blue_iface_t               *iface;
    struct usbd_device         *device;
    usbd_device_descriptor_t   *device_desc;
    struct usbd_desc_node      *conf;

    /* We want to ignore the isochronous interface on iface 1 */
    if (instance->iface != 0)
        return;

    if (usbd_attach(conn, instance, sizeof(blue_iface_t), &device) != EOK)
        return;
    
    iface           = usbd_device_extra(device);
    iface->blue     = &BlueCtrl;
    iface->device   = device;
    iface->instance = *instance;
    iface->status   = 0;
//    iface->devnum   = -1;
    pthread_mutex_init(&iface->mutex, NULL);
    TAILQ_INIT(&iface->cmd_q);
    iface->num_hci_cmd_packets = 1;
    TAILQ_INIT(&iface->connections);
    TAILQ_INIT(&iface->inquiry_cache);
    TAILQ_INIT(&iface->callbacks);

    if (blue_parse_descriptors(iface) != EOK)
        goto fail_detach;

    iface->ep_int_buf = usbd_alloc(iface->ep_int_size);
    if (!iface->ep_int_buf)
        goto fail_detach;
    
    iface->ep_bulk_buf = usbd_alloc(iface->ep_bulk_size);
    if (!iface->ep_bulk_buf)
        goto fail_free_int_buf;

    iface->urb_cmd = usbd_alloc_urb(NULL);
    if (!iface->urb_cmd)
        goto fail_free_bulk_buf;
    iface->urb_int = usbd_alloc_urb(NULL);
    if (!iface->urb_int)
        goto fail_free_bulk_buf;
    iface->urb_bulk_in = usbd_alloc_urb(NULL);
    if (!iface->urb_bulk_in)
        goto fail_free_bulk_buf;
    iface->urb_bulk_out = usbd_alloc_urb(NULL);
    if (!iface->urb_bulk_out)
        goto fail_free_bulk_buf;

    iofunc_attr_lock(&BlueCtrl.attr);
    TAILQ_INSERT_HEAD(&BlueCtrl.interfaces, iface, dlink);
    iofunc_attr_unlock(&BlueCtrl.attr);

    device_desc = usbd_device_descriptor(iface->device, &conf);
    if (!device_desc) {
        SLOG(_SLOG_INFO, "bluetooth: Initialized unknown Bluetooth module");
    }
    else {
        char str[256];
        
        str[0] = '\0';
        if (device_desc->iManufacturer) {
            sprintf(str, "%s ",
                    usbd_string(device, device_desc->iManufacturer, 0));
        }

        if (device_desc->iProduct)
            SLOG(_SLOG_INFO, "bluetooth: Initialized %s%s", str,
                    usbd_string(device, device_desc->iProduct, 0));
        else
            SLOG(_SLOG_INFO, "bluetooth: Initialized %sBluetooth module (%04x:%04x)",
                    str, device_desc->idVendor, device_desc->idProduct);
    }

    pthread_sleepon_lock();
    pthread_sleepon_signal(&BlueCtrl.interfaces);
    pthread_sleepon_unlock();
    return;

fail_free_bulk_buf:
    usbd_free(iface->ep_bulk_buf);
fail_free_int_buf:
    usbd_free(iface->ep_int_buf);
fail_detach:
    usbd_detach(device);
}

void blue_removal(struct usbd_connection * conn, usbd_device_instance_t *instance)
{
	struct usbd_device  *device;
    blue_iface_t        *iface;

    SLOG(_SLOG_INFO, "bluetooth: Device removed from iface %d", instance->iface);

	if((device = usbd_device_lookup( conn, instance ))) {
        iface = usbd_device_extra(device);
        blue_iface_free(iface);
	}
}

int blue_usb_start_driver (blue_t * blue, int argc, char **argv)
{
    int status;
    struct usbd_connection *connection;
    char opt;
    usbd_funcs_t        funcs    = {
        _USBDI_NFUNCS, blue_insert, blue_removal, NULL
    };
    
    /* matches the standard bluetooth class of devices */
    usbd_device_ident_t interest_class = {
        USBD_CONNECT_WILDCARD, USBD_CONNECT_WILDCARD, 0xE0, 0x01, 0x01
    };
    /* matches the Ericsson bluetooth chipset */
    usbd_device_ident_t interest_ericsson = {
        0x0BDB, 0x1002, 0xFF, 0xFF, 0xFF
    };

    usbd_connect_parm_t parm_class      = {
        NULL, USB_VERSION, USBD_VERSION, 0, argc, argv, 0,
        &interest_class,    &funcs, USBD_CONNECT_WAIT
    };
    usbd_connect_parm_t parm_ericsson   = {
        NULL, USB_VERSION, USBD_VERSION, 0, argc, argv, 0,
        &interest_ericsson, &funcs, USBD_CONNECT_WAIT
    };

    blue->verbose = 0;
    while ((opt = getopt(argc, argv, "vs:w:")) != -1) {
        switch (opt) {
            case 'v':
                blue->verbose++;
                break;
            case 's':
                parm_class.path    = optarg;
                parm_ericsson.path = optarg;
                break;
            case 'w':
                parm_class.connect_wait    = strtol(optarg, 0, 0);
                parm_ericsson.connect_wait = parm_class.connect_wait;
                break;
            default:
                break;
        }
    }
    if ((status = usbd_connect(&parm_class, &connection)) != EOK) {
        SLOG(_SLOG_ERROR, "bluetooth: USB failed to connect: %s", strerror(status));
        return status;
    }
    if ((status = usbd_connect(&parm_ericsson, &connection)) != EOK) {
        SLOG(_SLOG_ERROR, "bluetooth: USB failed to connect: %s", strerror(status));
        return status;
    }

    return EOK;
}

