/**
 * @file firmware.c
 *
 * Code to load firmware at initialization time for devices that
 * require it.  In this case, the Broadcom BCM2033 Bluetooth
 * device requires loading of firmware over the USB bus.
 *
 * Revision History:
 *   2003 Feb 4   David Moore      Initial version.
 */

#include <stdio.h>
#include "blue_driver.h"

static int blue_bulk_load(blue_iface_t * iface, char * path)
{
    char * buf;
    int len;
    FILE * file;

    buf = usbd_alloc(4096);
    if (!buf) {
        SLOG(_SLOG_CRITICAL, "bluetooth: Firmware load out of memory");
        return -1;
    }

    if (!(file = fopen(path, "r"))) {
        SLOG(_SLOG_ERROR, "bluetooth: Can't open firmware %s", path);
        return -1;
    }

    while (1) {
        len = fread(buf, 1, 4096, file);
        if (len < 0) {
            SLOG(_SLOG_ERROR, "bluetooth: Read failed loading firmware");
            goto error_abort;
        }
        if (!len)
            break;

        if (blue_bulk_io(iface, iface->ep_bulk_out, URB_DIR_OUT, buf, len)) {
            SLOG(_SLOG_ERROR, "bluetooth: Bulk transfer failed loading firmware");
            goto error_abort;
        }
    }
    fclose(file);
    usbd_free(buf);
    return 0;

error_abort:
    fclose(file);
    usbd_free(buf);
    return -1;
}

#define BROADCOM_VENDOR    0x0A5C
#define BROADCOM_PRODUCT   0x2033

int blue_load_firmware(blue_iface_t * iface)
{
    unsigned char * buf;
    usbd_device_descriptor_t * device_desc;
    struct usbd_desc_node * conf;

    device_desc = usbd_device_descriptor(iface->device, &conf);
    if (!device_desc)
        return -1;

    if (device_desc->idVendor != BROADCOM_VENDOR)
        return 0;
    if (device_desc->idProduct != BROADCOM_PRODUCT)
        return 0;

    SLOG(_SLOG_INFO, "bluetooth: Detected Broadcom chipset, uploading firmware...");

    buf = usbd_alloc(8);
    if (!buf) {
        SLOG(_SLOG_CRITICAL, "bluetooth: Out of memory loading firmware");
        return -1;
    }

    /* Load the mini-driver */
    if (blue_bulk_load(iface, "/etc/bluetooth/BCM2033-MD.hex"))
        return -1;

    usleep(10);
    
    buf[0] = '#';
    if (blue_bulk_io(iface, iface->ep_bulk_out, URB_DIR_OUT, buf, 1)) {
        SLOG(_SLOG_ERROR, "bluetooth: Bulk I/O failed checking firmware");
        return -1;
    }
    
    /* Confirm that the load was successful */
    if (blue_intr_in_sync(iface)) {
        SLOG(_SLOG_ERROR, "bluetooth: Mini-driver interrupt I/O failed");
        return -1;
    }
    if (iface->ep_int_buf[0] != '#') {
        SLOG(_SLOG_ERROR, "bluetooth: Mini-driver failed to load");
        return -1;
    }
    
    /* Load the firmware image */
    if (blue_bulk_load(iface, "/etc/bluetooth/BCM2033-FW.bin"))
        return -1;

    /* Confirm that the load was successful */
    if (blue_intr_in_sync(iface)) {
        SLOG(_SLOG_ERROR, "bluetooth: Firmware interrupt I/O failed");
        return -1;
    }
    if (iface->ep_int_buf[0] != '.') {
        SLOG(_SLOG_ERROR, "bluetooth: Firmware image failed to load 0x%02x",
                iface->ep_int_buf[0]);
        return -1;
    }

    SLOG(_SLOG_INFO, "bluetooth: Successfully loaded bluetooth firmware.");
    return 1;
}
