/**
 * @file hci.c
 *
 * This layer sits under the L2CAP layer and manages the Host Controller
 * Interface (HCI) to the Bluetooth device.  HCI provides a set of
 * commands and events that are sent and received, respectively, from
 * the device.  In addition, data is sent and received using either
 * ACL or SCO packets, which are formatted according the HCI specs and
 * sent/received from the Bluetooth device.
 *
 * In this file, commands, events, and data packets are constructed and
 * deconstructed as received from lower or higher layers.  Commands
 * and data packets to be sent are queued, and another thread in
 * blue.c handles the transmission to the device.  Received events and
 * data packets are processed and then passed to a higher layer (L2CAP)
 * if necessary.
 *
 * Revision History:
 *   2003 Feb 4   David Moore      Initial version.
 */

#include <stdio.h>
#include "blue_driver.h"
#include <bluetooth/hci.h>


int hci_send_command(blue_iface_t * iface, int ogf, int ocf,
        void * param, int param_len)
{
    databuf_t * cmd;

    cmd = (databuf_t *) malloc (sizeof(databuf_t) + 3 + param_len);
    if (!cmd)
        return ENOMEM;

    cmd->len = 3 + param_len;
    *((unsigned short *)cmd->data) = HCI_OPCODE(ogf, ocf);
    cmd->data[2] = param_len;
    if (param) {
        memcpy(cmd->data+3, param, param_len);
    }

    pthread_mutex_lock(&iface->mutex);
    TAILQ_INSERT_TAIL(&iface->cmd_q, cmd, dlink);
    pthread_mutex_unlock(&iface->mutex);

    pthread_sleepon_lock();
    pthread_sleepon_signal(&iface->cmd_q);
    pthread_sleepon_unlock();

    return EOK;
}

static callback_entry_t * find_callback(blue_iface_t * iface, unsigned short opcode)
{
    callback_entry_t * entry;

    entry = TAILQ_FIRST(&iface->callbacks);
    while (entry) {
        if (entry->opcode == opcode)
            return entry;
        entry = TAILQ_NEXT(entry, dlink);
    }
    return NULL;
}

static callback_entry_t * find_callback_by_event(blue_iface_t * iface, unsigned char event)
{
    callback_entry_t * entry;

    entry = TAILQ_FIRST(&iface->callbacks);
    while (entry) {
        if (entry->event == event)
            return entry;
        entry = TAILQ_NEXT(entry, dlink);
    }
    return NULL;
}

int hci_queue_callback(blue_iface_t * iface, int ogf, int ocf, int event,
        reply_info_t reply_info, hci_status_callback_t status_callback,
        hci_event_callback_t event_callback)
{
    unsigned short opcode = HCI_OPCODE(ogf, ocf);
    callback_entry_t * entry;

    if (find_callback(iface, opcode))
        return EINVAL;

    entry = (callback_entry_t *) malloc(sizeof(callback_entry_t));
    if (!entry)
        return ENOMEM;

    entry->opcode           = opcode;
    entry->reply_info       = reply_info;
    entry->event            = event;
    entry->status_callback  = status_callback;
    entry->event_callback   = event_callback;

    TAILQ_INSERT_TAIL(&iface->callbacks, entry, dlink);
    
    return EOK;
}

int hci_send_acl(hci_conn_t * conn, unsigned int flags,
        unsigned char * data, unsigned int len)
{
    blue_iface_t * iface = conn->iface;
    databuf_t * buf;
    hci_acldata_t * acl;

    buf = (databuf_t *) malloc(len + sizeof(hci_acldata_t) + sizeof(databuf_t));
    if (!buf)
        return ENOMEM;

    buf->len = sizeof(hci_acldata_t) + len;
    
    acl = (hci_acldata_t *) buf->data;
    acl->handle = 0;
    if (flags & HCI_ACL_CONT)
        acl->boundary = ACL_CONT;
    else
        acl->boundary = ACL_FIRST;
    
    if (flags & HCI_ACTIVE_BCAST)
        acl->broadcast = ACL_ACTIVE_BCAST;
    else if (flags & HCI_PNET_BCAST)
        acl->broadcast = ACL_PNET_BCAST;
    else
        acl->broadcast = ACL_NO_BCAST;
    
    acl->len = ENDIAN_LE16(len);
    memcpy(acl->data, data, len);

    pthread_mutex_lock(&conn->acl_mutex);
    TAILQ_INSERT_TAIL(&conn->acl_q, buf, dlink);
    pthread_mutex_unlock(&conn->acl_mutex);

    pthread_sleepon_lock();
    pthread_sleepon_signal(&iface->num_hci_acl_packets);
    pthread_sleepon_unlock();

    return EOK;
}

static char bdstr[24];

char * hci_bdaddr_str(bdaddr_t bd)
{
    int i;

    for (i=0; i < sizeof(bd.b); i++) {
        sprintf(bdstr + 3*i, "%02x%s", bd.b[sizeof(bd.b)-1-i],
                (i == sizeof(bd.b)-1)?"":":");
    }
    return bdstr;
}

#if 0
int hci_cmp_bdaddr(bdaddr_t bd1, bdaddr_t bd2)
{
    int i;
    int match = 0;

    for (i=0; i < sizeof(bd1.b); i++) {
        if (bd1.b[i] != bd2.b[i])
            match = 1;
    }
    return match;
}
#endif

hci_conn_t * hci_get_conn_by_handle(blue_iface_t * iface, unsigned int handle)
{
    hci_conn_t * conn;

    pthread_mutex_lock(&iface->mutex);
    conn = TAILQ_FIRST(&iface->connections);
    while (conn) {
        if (conn->handle == handle) {
            pthread_mutex_unlock(&iface->mutex);
            return conn;
        }
        conn = TAILQ_NEXT(conn, dlink);
    }
    pthread_mutex_unlock(&iface->mutex);
    return NULL;
}

hci_conn_t * hci_get_conn_by_bdaddr(blue_t * blue, blue_iface_t * iface, bdaddr_t bdaddr)
{
    hci_conn_t * conn;

    if (!iface) {
        blue_iface_t * iface;

        iofunc_attr_lock(&blue->attr);
        iface = TAILQ_FIRST(&blue->interfaces);
        while (iface) {
            conn = hci_get_conn_by_bdaddr(blue, iface, bdaddr);
            if (conn) {
                iofunc_attr_unlock(&blue->attr);
                return conn;
            }
            iface = TAILQ_NEXT(iface, dlink);
        }
        iofunc_attr_unlock(&blue->attr);
        return NULL;
    }

    pthread_mutex_lock(&iface->mutex);
    conn = TAILQ_FIRST(&iface->connections);
    while (conn) {
        if (bacmp(&conn->bdaddr, &bdaddr) == 0) {
            pthread_mutex_unlock(&iface->mutex);
            return conn;
        }
        conn = TAILQ_NEXT(conn, dlink);
    }
    pthread_mutex_unlock(&iface->mutex);
    return NULL;
}

inq_res_t * hci_find_inquiry_result(blue_iface_t * iface, bdaddr_t bdaddr)
{
    inq_entry_t * entry;

    entry = TAILQ_FIRST(&iface->inquiry_cache);
    while (entry) {
        if (!bacmp(&entry->i.bdaddr, &bdaddr)) {
            return &entry->i;
        }
        entry = TAILQ_NEXT(entry, dlink);
    }

    return NULL;
}


blue_iface_t * hci_get_default_iface(blue_t * blue)
{
    blue_iface_t * iface;

    iofunc_attr_lock(&blue->attr);
    iface = TAILQ_FIRST(&blue->interfaces);
    iofunc_attr_unlock(&blue->attr);
    return iface;
}

blue_iface_t * hci_get_iface_by_bdaddr(blue_t * blue, bdaddr_t bdaddr)
{
    blue_iface_t * iface;

    iofunc_attr_lock(&blue->attr);
    iface = TAILQ_FIRST(&blue->interfaces);
    while (iface) {
        if (!bacmp(&iface->bdaddr, &bdaddr) ||
                !bacmp(&bdaddr, BDADDR_ANY)) {
            iofunc_attr_unlock(&blue->attr);
            return iface;
        }
        iface = TAILQ_NEXT(iface, dlink);
    }

    iofunc_attr_unlock(&blue->attr);
    return NULL;
}

static hci_conn_t * alloc_conn(blue_iface_t * iface, bdaddr_t bdaddr)
{
    hci_conn_t * conn;

    conn = (hci_conn_t *) malloc(sizeof(hci_conn_t));
    if (!conn)
        return NULL;

    conn->iface = iface;
    conn->handle = 0;
    conn->bdaddr = bdaddr;
    pthread_mutex_init(&conn->acl_mutex, NULL);
    TAILQ_INIT(&conn->acl_q);
    conn->acl_pending = 0;

    pthread_mutex_lock(&iface->mutex);
    TAILQ_INSERT_HEAD(&iface->connections, conn, dlink);
    pthread_mutex_unlock(&iface->mutex);

    conn->recv_buf_pos = 0;

    return conn;
}

static void free_conn(blue_iface_t * iface, hci_conn_t * conn)
{
    databuf_t * acl;

    pthread_mutex_lock(&conn->acl_mutex);
    acl = TAILQ_FIRST(&conn->acl_q);
    while (acl) {
        TAILQ_REMOVE(&conn->acl_q, acl, dlink);
        free(acl);
        acl = TAILQ_FIRST(&conn->acl_q);
    }
    pthread_mutex_unlock(&conn->acl_mutex);

    pthread_mutex_lock(&iface->mutex);
    TAILQ_REMOVE(&iface->connections, conn, dlink);
    pthread_mutex_unlock(&iface->mutex);

    free(conn);
}

void hci_free(blue_iface_t * iface)
{
    hci_conn_t * conn;
    cmd_disconn_t disconn;
    inq_entry_t * entry;

    pthread_mutex_lock(&iface->mutex);
    while (1) {
        entry = TAILQ_FIRST(&iface->inquiry_cache);
        if (!entry)
            break;

        TAILQ_REMOVE(&iface->inquiry_cache, entry, dlink);
        free(entry);
    }
    pthread_mutex_unlock(&iface->mutex);

    while (1) {
        pthread_mutex_lock(&iface->mutex);
        conn = TAILQ_FIRST(&iface->connections);
        pthread_mutex_unlock(&iface->mutex);
        if (!conn)
            break;

        if (conn->handle) {
            disconn.handle = ENDIAN_LE16(conn->handle);
            disconn.reason = HCI_ERROR_USER_ENDED;
            hci_send_command(iface, HCI_OGF_LINK_CONTROL, HCI_DISCONN,
                    &disconn, sizeof(disconn));
        }
        free_conn(iface, conn);
    }
    usleep(200000);
}


hci_conn_t * hci_connect(blue_iface_t * iface, bdaddr_t bdaddr)
{
    hci_conn_t * conn;
    cmd_create_conn_t cmd;
    inq_res_t * inq;

    if ((conn = hci_get_conn_by_bdaddr(iface->blue, iface, bdaddr)))
        return conn;

    conn = alloc_conn(iface, bdaddr);
    if (!conn)
        return NULL;
    
    cmd.bdaddr            = bdaddr;
    cmd.pkt_type          = ENDIAN_LE16(PKT_TYPE_DEFAULT);
    cmd.pg_scan_rep_mode  = 0;
    cmd.pg_scan_mode      = 0;
    cmd.clk_offset        = 0;
    cmd.allow_role_switch = 1;
    
    pthread_mutex_lock(&iface->mutex);
    inq = hci_find_inquiry_result(iface, bdaddr);
    if (inq) {
        SLOG(_SLOG_DEBUG1, "bluetooth: Connecting with saved offset 0x%x",
                ENDIAN_LE16(inq->clock_offset));
        cmd.pg_scan_rep_mode = inq->page_scan_rep_mode;
        cmd.pg_scan_mode     = inq->page_scan_mode;
        cmd.clk_offset       = ENDIAN_LE16(ENDIAN_LE16(inq->clock_offset) | 0x80);
    }
    pthread_mutex_unlock(&iface->mutex);
    
    hci_send_command(iface, HCI_OGF_LINK_CONTROL, HCI_CREATE_CONN,
            &cmd, sizeof(cmd));

    return conn;
}

hci_conn_t * hci_sched_next_acl(blue_iface_t * iface, unsigned * num_packets)
{
    hci_conn_t * conn;
    hci_conn_t * min_conn = NULL;
    unsigned int min_pend = ~0;
    unsigned int total_pend = 0;

    pthread_mutex_lock(&iface->mutex);
    conn = TAILQ_FIRST(&iface->connections);
    while (conn) {
        pthread_mutex_lock(&conn->acl_mutex);
        if (TAILQ_FIRST(&conn->acl_q) && conn->handle
                && conn->acl_pending < iface->num_hci_acl_packets) {

            total_pend++;
            if (conn->acl_pending < min_pend) {
                min_pend = conn->acl_pending;
                min_conn = conn;
            }
        }
        pthread_mutex_unlock(&conn->acl_mutex);

        conn = TAILQ_NEXT(conn, dlink);
    }

    if (min_conn) {
        int n = iface->num_hci_acl_packets / total_pend;
        if (n >= 1)
            *num_packets = n;
        else
            *num_packets = 1;
    }
    pthread_mutex_unlock(&iface->mutex);

    return min_conn;
}

static char errstr[256];

static char * hci_error_str(unsigned char code)
{
    switch(code) {
        default:
            sprintf(errstr, "Error code 0x%02x", code);
            break;
    }
    return errstr;
}

static void add_inquiry_result(blue_iface_t * iface, inq_res_t * result)
{
    inq_entry_t * entry;
    inq_res_t * oldres;

    pthread_mutex_lock(&iface->mutex);

    oldres = hci_find_inquiry_result(iface, result->bdaddr);
    if (oldres) {
        SLOG(_SLOG_DEBUG1, "bluetooth: Updating inquiry cache for %s",
                hci_bdaddr_str(result->bdaddr));
        memcpy(oldres, result, sizeof(inq_res_t));
        pthread_mutex_unlock(&iface->mutex);
        return;
    }

    entry = (inq_entry_t *) malloc(sizeof(inq_entry_t));
    if (entry) {
        SLOG(_SLOG_DEBUG1, "bluetooth: New inquiry cache for %s",
                hci_bdaddr_str(result->bdaddr));
        memcpy(&entry->i, result, sizeof(inq_res_t));
        TAILQ_INSERT_HEAD(&iface->inquiry_cache, entry, dlink);
    }

    pthread_mutex_unlock(&iface->mutex);
}

static int event_inquiry_result(blue_iface_t * iface, evt_inquiry_result * evt)
{
    int num = evt->num_responses;
    int i;
    inq_res_t entry[num];
    unsigned char * a = evt->data;

    for (i=0; i<num; i++) {
        entry[i].bdaddr = *(bdaddr_t *)a;
        a += sizeof(bdaddr_t);
    }

    for (i=0; i<num; i++) {
        entry[i].page_scan_rep_mode = *a;
        a++;
    }

    for (i=0; i<num; i++) {
        entry[i].page_scan_period_mode = *a;
        a++;
    }

    for (i=0; i<num; i++) {
        entry[i].page_scan_mode = *a;
        a++;
    }

    for (i=0; i<num; i++) {
        entry[i].class_of_device = *(unsigned int *)a & 0xffffff;
        a += 3;
    }

    for (i=0; i<num; i++) {
        entry[i].clock_offset = *(unsigned short *)a;
        a += 2;
    }

    for (i=0; i<num; i++) {
        add_inquiry_result(iface, entry+i);
    }

    return EOK;
}

static int event_conn_complete(blue_iface_t * iface, evt_conn_complete * evt)
{
    hci_conn_t * conn;

    conn = hci_get_conn_by_bdaddr(iface->blue, iface, evt->bdaddr);
    if (!conn) {
        SLOG(_SLOG_NOTICE, "bluetooth: Unexpected connection complete");
        return EOK;
    }

    if (evt->status != 0) {
        SLOG(_SLOG_DEBUG1, "bluetooth: Connection failed: %s",
                hci_error_str(evt->status));
        free_conn(iface, conn);
        return EOK;
    }

    if (evt->link_type != 0x1) {
        SLOG(_SLOG_DEBUG1, "bluetooth: Non-ACL connection ignored");
        return EOK;
    }

    if (evt->encryption_mode != 0) {
        SLOG(_SLOG_NOTICE, "bluetooth: Encrypted connection ignored");
        return EOK;
    }

    if (conn->handle) {
        SLOG(_SLOG_NOTICE, "bluetooth: Conn complete on existing connection");
    }
    conn->handle = ENDIAN_LE16(evt->handle);

    SLOG(_SLOG_DEBUG1, "bluetooth: Connection complete from %s",
            hci_bdaddr_str(evt->bdaddr));
    pthread_sleepon_lock();
    pthread_sleepon_signal(&iface->num_hci_acl_packets);
    pthread_sleepon_unlock();

    return EOK;
}

static int event_conn_request(blue_iface_t * iface, evt_conn_request * evt)
{
    hci_conn_t * conn;
    cmd_accept_conn_req_t accept;

    /* We won't accept non-ACL connections */
    if (evt->link_type != 0x1) {
        cmd_reject_conn_req_t reject;

        SLOG(_SLOG_DEBUG1, "bluetooth: Rejected non-ACL connection");
        reject.bdaddr = evt->bdaddr;
        reject.reason = HCI_ERROR_LIMITED_RESOURCES;
        hci_send_command(iface, HCI_OGF_LINK_CONTROL, HCI_REJECT_CONN_REQ,
                &reject, sizeof(reject));
        return EOK;
    }

    conn = hci_get_conn_by_bdaddr(iface->blue, iface, evt->bdaddr);
    if (!conn) {
        conn = alloc_conn(iface, evt->bdaddr);
        if (!conn)
            return ENOMEM;
    }
    else {
        SLOG(_SLOG_NOTICE, "bluetooth: Conn request on existing connection");
    }

    SLOG(_SLOG_DEBUG1, "bluetooth: Accepted connection from %s",
            hci_bdaddr_str(evt->bdaddr));
    accept.bdaddr = evt->bdaddr;
    accept.role   = 0x1;
    hci_send_command(iface, HCI_OGF_LINK_CONTROL, HCI_ACCEPT_CONN_REQ,
            &accept, sizeof(accept));

    return EOK;
}

static int event_disconn_complete(blue_iface_t * iface, evt_disconn_complete * evt)
{
    hci_conn_t * conn;

    if (evt->status != 0) {
        SLOG(_SLOG_NOTICE, "bluetooth: Disconnection failed: %s",
                hci_error_str(evt->status));
        return EOK;
    }
    
    conn = hci_get_conn_by_handle(iface, ENDIAN_LE16(evt->handle));
    if (!conn) {
        return EOK;
    }

    free_conn(iface, conn);
    return EOK;
}

static int event_command_complete(blue_iface_t * iface, evt_command_complete * evt)
{
    cmdcmp_read_buffer_size_t * crbs = NULL;
    cmdcmp_read_bdaddr_t * crba = NULL;
    callback_entry_t * cbk;
        
    pthread_mutex_lock(&iface->mutex);
    iface->num_hci_cmd_packets = evt->num_hci_cmd_packets;

    pthread_sleepon_lock();
    pthread_sleepon_signal(&iface->num_hci_cmd_packets);
    pthread_sleepon_unlock();

    if ((cbk = find_callback(iface, evt->opcode))) {
        TAILQ_REMOVE(&iface->callbacks, cbk, dlink);
        pthread_mutex_unlock(&iface->mutex);
        cbk->event_callback(iface, cbk->reply_info, evt->param);
        free(cbk);
        return EOK;
    }
    pthread_mutex_unlock(&iface->mutex);

    switch (evt->opcode) {
        case HCI_OPCODE(HCI_OGF_INFO, HCI_READ_BUFFER_SIZE):
            crbs = (cmdcmp_read_buffer_size_t *) evt->param;
            if (crbs->status != 0) {
                SLOG(_SLOG_ERROR, "bluetooth: Read buffer size failed: %s",
                        hci_error_str(crbs->status));
                break;
            }
            pthread_mutex_lock(&iface->mutex);
            iface->num_hci_acl_packets   = crbs->max_acl_packets;
            iface->max_acl_data_len      = crbs->max_acl_len;
            pthread_mutex_unlock(&iface->mutex);
            break;

        case HCI_OPCODE(HCI_OGF_HC_AND_BASEBAND, HCI_HOST_BUFFER_SIZE):
        case HCI_OPCODE(HCI_OGF_HC_AND_BASEBAND, HCI_WRITE_SCAN_ENABLE):
        case HCI_OPCODE(HCI_OGF_LINK_CONTROL, HCI_PER_INQUIRY):
        case HCI_OPCODE(HCI_OGF_LINK_CONTROL, HCI_EXIT_PER_INQUIRY):
            if (evt->param[0] != 0) {
                SLOG(_SLOG_ERROR, "bluetooth: Command 0x%04x failed: %s",
                        ENDIAN_LE16(evt->opcode), hci_error_str(evt->param[0]));
            }
            break;

        case HCI_OPCODE(HCI_OGF_INFO, HCI_READ_BDADDR):
            crba = (cmdcmp_read_bdaddr_t *) evt->param;
            if (crba->status != 0) {
                SLOG(_SLOG_ERROR, "bluetooth: Read BDADDR failed: %s",
                        hci_error_str(crba->status));
                break;
            }
            iface->bdaddr = crba->bdaddr;
            iface->status |= BLUE_READY;
            SLOG(_SLOG_INFO, "bluetooth: Added interface %s",
                    hci_bdaddr_str(iface->bdaddr));
            break;

        default:
            SLOG(_SLOG_INFO, "bluetooth: Unhandled command complete 0x%04x",
                    ENDIAN_LE16(evt->opcode));
            break;
    }

    return EOK;
}

static int event_command_status(blue_iface_t * iface, evt_command_status * evt)
{
    pthread_mutex_lock(&iface->mutex);
    iface->num_hci_cmd_packets = evt->num_hci_cmd_packets;

    pthread_sleepon_lock();
    pthread_sleepon_signal(&iface->num_hci_cmd_packets);
    pthread_sleepon_unlock();

    if (evt->status != 0) {
        callback_entry_t * cbk;

        if ((cbk = find_callback(iface, evt->opcode))) {
            TAILQ_REMOVE(&iface->callbacks, cbk, dlink);
            pthread_mutex_unlock(&iface->mutex);
            cbk->status_callback(iface, cbk->reply_info, evt->status);
            free(cbk);
            return EOK;
        }

        SLOG(_SLOG_NOTICE, "bluetooth: Command 0x%04x status error: %s",
                ENDIAN_LE16(evt->opcode), hci_error_str(evt->status));
    }
    pthread_mutex_unlock(&iface->mutex);

    return EOK;
}

static int event_num_comp_packets(blue_iface_t * iface, unsigned char * evt)
{
    int num_handles = evt[0];
    int total_sent = 0;
    int i;

    for (i=0; i<num_handles; i++) {
        unsigned short handle   = *(unsigned short *)(evt+1+(i<<1));
        unsigned short num_pkts = *(unsigned short *)(evt+1+((i+num_handles)<<1));
        hci_conn_t * conn;

        total_sent += num_pkts;

        if (!(conn = hci_get_conn_by_handle(iface, handle)))
            continue;

        pthread_mutex_lock(&conn->acl_mutex);
        conn->acl_pending -= num_pkts;
        pthread_mutex_unlock(&conn->acl_mutex);
    }

    pthread_mutex_lock(&iface->mutex);
    iface->num_hci_acl_packets += total_sent;
    pthread_mutex_unlock(&iface->mutex);

    pthread_sleepon_lock();
    pthread_sleepon_signal(&iface->num_hci_acl_packets);
    pthread_sleepon_unlock();

    return EOK;
}

static int event_page_scan_rep_mode_change(blue_iface_t * iface,
        evt_page_scan_rep_mode_change * evt)
{
    inq_res_t * result;

    pthread_mutex_lock(&iface->mutex);

    result = hci_find_inquiry_result(iface, evt->bdaddr);
    if (result) {
        result->page_scan_rep_mode = evt->mode;
        SLOG(_SLOG_DEBUG1, "bluetooth: %s changed page scan rep mode to R%d",
                hci_bdaddr_str(evt->bdaddr), evt->mode);
    }
    
    pthread_mutex_unlock(&iface->mutex);

    return EOK;
}

#define DISPATCH_EVENT(type,func) do {      \
    type * evt = (type *) header->param;    \
    if (len < sizeof(type)) {               \
        return EINVAL;                      \
    }                                       \
    func(iface, evt);                       \
} while (0)

int hci_event(blue_iface_t * iface, unsigned char * buf, _uint32 len)
{
    hci_event_t * header = (hci_event_t *) buf;
    callback_entry_t * cbk;

    if (len < sizeof(hci_event_t)) {
        SLOG(_SLOG_NOTICE, "bluetooth: Dropped malformed event of len %d", len);
        return EINVAL;
    }
    len -= sizeof(hci_event_t);

    pthread_mutex_lock(&iface->mutex);
    if ((cbk = find_callback_by_event(iface, header->event))) {
        TAILQ_REMOVE(&iface->callbacks, cbk, dlink);
        pthread_mutex_unlock(&iface->mutex);
        cbk->event_callback(iface, cbk->reply_info, header->param);
        free(cbk);
        return EOK;
    }
    pthread_mutex_unlock(&iface->mutex);

    switch (header->event) {
        case HCI_EVT_INQUIRY_COMPLETE:
            if (header->param[0] == 0)
                SLOG(_SLOG_DEBUG1, "bluetooth: Inquiry completed successfully");
            else
                SLOG(_SLOG_NOTICE, "bluetooth: Inquiry failed: %s",
                        hci_error_str(header->param[0]));
            break;
        case HCI_EVT_INQUIRY_RESULT:
            DISPATCH_EVENT(evt_inquiry_result, event_inquiry_result);
            break;
        case HCI_EVT_CONN_COMPLETE:
            DISPATCH_EVENT(evt_conn_complete, event_conn_complete);
            break;
        case HCI_EVT_CONN_REQUEST:
            DISPATCH_EVENT(evt_conn_request, event_conn_request);
            break;
        case HCI_EVT_DISCONN_COMPLETE:
            DISPATCH_EVENT(evt_disconn_complete, event_disconn_complete);
            break;
        case HCI_EVT_COMMAND_COMPLETE:
            DISPATCH_EVENT(evt_command_complete, event_command_complete);
            break;
        case HCI_EVT_COMMAND_STATUS:
            DISPATCH_EVENT(evt_command_status, event_command_status);
            break;
        case HCI_EVT_NUM_COMP_PACKETS:
            DISPATCH_EVENT(unsigned char, event_num_comp_packets);
        case HCI_EVT_MAX_SLOTS_CHANGE:
            break; /* ignore it */
        case HCI_EVT_PAGE_SCAN_REP_MODE_CHANGE:
            DISPATCH_EVENT(evt_page_scan_rep_mode_change, event_page_scan_rep_mode_change);
            break;
        default:
            SLOG(_SLOG_NOTICE, "bluetooth: Ignoring unknown event 0x%02x",
                    header->event);
            break;
    }
    return EOK;
}

int hci_aclrecv(blue_iface_t * iface, unsigned char * buf, _uint32 len)
{
    hci_acldata_t * acl = (hci_acldata_t *) buf;
    hci_conn_t * conn;
    int acl_len = ENDIAN_LE16(acl->len);
    
    if (acl_len + 4 != len) {
        SLOG(_SLOG_NOTICE, "bluetooth: Dropped malformed ACL packet");
        return EOK;
    }

    conn = hci_get_conn_by_handle(iface, ENDIAN_LE16(acl->handle));

    if (conn) {
        unsigned short l2cap_len;

        if (acl->boundary == ACL_FIRST && conn->recv_buf_pos != 0) {
            SLOG(_SLOG_NOTICE, "bluetooth: Dropped incomplete L2CAP packet");
            conn->recv_buf_pos = 0;
        }

        if (acl->boundary == ACL_CONT && conn->recv_buf_pos == 0) {
            SLOG(_SLOG_NOTICE, "bluetooth: Missing start of L2CAP packet");
            return EOK;
        }
        
        if (conn->recv_buf_pos + acl_len > L2CAP_MTU) {
            SLOG(_SLOG_NOTICE, "bluetooth: Dropped packet that exceeds MTU");
            conn->recv_buf_pos = 0;
            return EOK;
        }

        memcpy(&conn->recv_buf[conn->recv_buf_pos], acl->data, acl_len);
        conn->recv_buf_pos += ENDIAN_LE16(acl->len);

        l2cap_len = ENDIAN_LE16(*(unsigned short *)conn->recv_buf);
        if (l2cap_len + 4 == conn->recv_buf_pos) {
            l2cap_recv_packet(conn, conn->recv_buf, l2cap_len);
            conn->recv_buf_pos = 0;
        }
    }
    else {
        SLOG(_SLOG_NOTICE, "bluetooth: Dropped packet w/ unknown handle 0x%04x",
                ENDIAN_LE16(acl->handle));
    }
    
    return EOK;
}

