/**
 * @file iface.c
 *
 * This code provides a device file /dev/blue/hci* for each interface.
 * User-space code can send low-level commands to the driver through
 * this device file that would not work through the normal socket mechanism,
 * such as retrieving inquiry results.
 *
 * Revision History:
 *   2003 Feb 26   David Moore      Initial version.
 */

#include <stdio.h>

/* These overrides must be defined before iofunc.h is included */
struct ocb_iface;
#define IOFUNC_OCB_T    struct ocb_iface

#include "blue_driver.h"

struct ocb_iface {
    iofunc_ocb_t            hdr;
};

static int iface_io_inquiry(blue_iface_t * iface, resmgr_context_t * ctp,
        io_devctl_t * msg)
{
    iov_t * iov;
    int num_responses;
    inq_entry_t * entry;
    dcmd_blue_inquiry_t inq;
    int i;

    if (msg->i.nbytes < sizeof(dcmd_blue_inquiry_t))
        return EINVAL;

    resmgr_msgread(ctp, &inq, sizeof(dcmd_blue_inquiry_t), sizeof(msg->i));

    iov = (iov_t *) malloc((inq.num_responses+2) * sizeof(iov_t));
    if (!iov)
        return ENOMEM;

    SETIOV(&iov[0], &msg->o, sizeof(msg->o));
    SETIOV(&iov[1], &num_responses, sizeof(int));

    i = 0;
    entry = TAILQ_FIRST(&iface->inquiry_cache);
    while (entry) {
        if (i >= inq.num_responses)
            break;
        SETIOV(&iov[i+2], &entry->i, sizeof(inq_res_t));
        entry = TAILQ_NEXT(entry, dlink);
        i++;
    }
    num_responses = i;

    memset(&msg->o, 0, sizeof(msg->o));
    msg->o.nbytes = sizeof(dcmd_blue_inquiry_t) + num_responses * sizeof(inq_res_t);

    resmgr_msgwritev(ctp, iov, num_responses+2, 0);
    free(iov);
    
    return EOK;
}

static int callback_remote_name_failure(blue_iface_t * iface,
        reply_info_t reply_info, unsigned char status)
{
    MsgError(reply_info.rcvid, EIO);
    return EOK;
}

static int callback_remote_name_complete(blue_iface_t * iface,
        reply_info_t reply_info, unsigned char * result)
{
    cmdcmp_remote_name_req_t * req = (cmdcmp_remote_name_req_t *) result;

    if (req->status != 0) {
        MsgError(reply_info.rcvid, EIO);
        return EOK;
    }

    MsgWrite(reply_info.rcvid, req->remote_name, 248, reply_info.offset);
    MsgReply(reply_info.rcvid, EOK, NULL, 0);
    return EOK;
}

static int iface_io_remote_name(blue_iface_t * iface, resmgr_context_t * ctp,
        io_devctl_t * msg)
{
    dcmd_blue_remote_name_t input;
    cmd_remote_name_req_t req;
    inq_res_t * inq;
    reply_info_t reply_info;

    if (msg->i.nbytes != sizeof(dcmd_blue_remote_name_t))
        return EINVAL;

    resmgr_msgread(ctp, &input, sizeof(dcmd_blue_remote_name_t), sizeof(msg->i));

    req.bdaddr            = input.bdaddr;
    req.pg_scan_rep_mode  = 0;
    req.pg_scan_mode      = 0;
    req.clk_offset        = 0;
    
    inq = hci_find_inquiry_result(iface, input.bdaddr);
    if (inq) {
        SLOG(_SLOG_DEBUG1, "bluetooth: Getting remote name with saved offset 0x%x",
                ENDIAN_LE16(inq->clock_offset));
        req.pg_scan_rep_mode = inq->page_scan_rep_mode;
        req.pg_scan_mode     = inq->page_scan_mode;
        req.clk_offset       = ENDIAN_LE16(ENDIAN_LE16(inq->clock_offset) | 0x80);
    }

    memset(&msg->o, 0, sizeof(msg->o));
    msg->o.nbytes = sizeof(dcmd_blue_remote_name_t);
    resmgr_msgwrite(ctp, &msg->o, sizeof(msg->o), 0);

    reply_info.rcvid = ctp->rcvid;
    reply_info.offset = sizeof(msg->o) + sizeof(bdaddr_t);

    if (hci_queue_callback(iface, HCI_OGF_LINK_CONTROL, HCI_REMOTE_NAME_REQ,
            HCI_EVT_REMOTE_NAME_REQ_COMPLETE, reply_info, callback_remote_name_failure,
            callback_remote_name_complete) != EOK)
        return EAGAIN;

    hci_send_command(iface, HCI_OGF_LINK_CONTROL, HCI_REMOTE_NAME_REQ,
            &req, sizeof(req));

    return _RESMGR_NOREPLY;
}

static int callback_change_name_complete(blue_iface_t * iface,
        reply_info_t reply_info, unsigned char * result)
{
    unsigned char status = *result;

    if (status != 0) {
        MsgError(reply_info.rcvid, EIO);
        return EOK;
    }

    MsgReply(reply_info.rcvid, EOK, NULL, 0);
    return EOK;
}

static int iface_io_change_name(blue_iface_t * iface, resmgr_context_t * ctp,
        io_devctl_t * msg)
{
    char name[248];
    int len = min(msg->i.nbytes, sizeof(name));
    reply_info_t reply_info;

    resmgr_msgread(ctp, name, len, sizeof(msg->i));
    if (len < sizeof(name))
        name[len] = '\0';
    else
        name[len-1] = '\0';

    reply_info.rcvid = ctp->rcvid;
    reply_info.offset = 0;

    if (hci_queue_callback(iface, HCI_OGF_HC_AND_BASEBAND, HCI_CHANGE_LOCAL_NAME,
            0, reply_info, NULL, callback_change_name_complete) != EOK)
        return EAGAIN;

    hci_send_command(iface, HCI_OGF_HC_AND_BASEBAND, HCI_CHANGE_LOCAL_NAME,
            name, sizeof(name));

    return _RESMGR_NOREPLY;
}

static int callback_local_name_complete(blue_iface_t * iface,
        reply_info_t reply_info, unsigned char * result)
{
    cmdcmp_read_local_name_t * req = (cmdcmp_read_local_name_t *) result;

    if (req->status != 0) {
        MsgError(reply_info.rcvid, EIO);
        return EOK;
    }

    MsgWrite(reply_info.rcvid, req->name, 248, reply_info.offset);
    MsgReply(reply_info.rcvid, EOK, NULL, 0);
    return EOK;
}

static int iface_io_local_name(blue_iface_t * iface, resmgr_context_t * ctp,
        io_devctl_t * msg)
{
    reply_info_t reply_info;

    memset(&msg->o, 0, sizeof(msg->o));
    msg->o.nbytes = 248;
    resmgr_msgwrite(ctp, &msg->o, sizeof(msg->o), 0);

    reply_info.rcvid = ctp->rcvid;
    reply_info.offset = sizeof(msg->o);

    if (hci_queue_callback(iface, HCI_OGF_HC_AND_BASEBAND, HCI_READ_LOCAL_NAME,
            0, reply_info, NULL, callback_local_name_complete) != EOK)
        return EAGAIN;

    hci_send_command(iface, HCI_OGF_HC_AND_BASEBAND, HCI_READ_LOCAL_NAME, NULL, 0);

    return _RESMGR_NOREPLY;
}

static int callback_get_bdaddr_complete(blue_iface_t * iface,
        reply_info_t reply_info, unsigned char * result)
{
    cmdcmp_read_bdaddr_t * req = (cmdcmp_read_bdaddr_t *) result;

    if (req->status != 0) {
        MsgError(reply_info.rcvid, EIO);
        return EOK;
    }

    MsgWrite(reply_info.rcvid, &req->bdaddr, sizeof(bdaddr_t), reply_info.offset);
    MsgReply(reply_info.rcvid, EOK, NULL, 0);
    return EOK;
}

static int iface_io_get_bdaddr(blue_iface_t * iface, resmgr_context_t * ctp,
        io_devctl_t * msg)
{
    reply_info_t reply_info;

    memset(&msg->o, 0, sizeof(msg->o));
    msg->o.nbytes = sizeof(bdaddr_t);
    resmgr_msgwrite(ctp, &msg->o, sizeof(msg->o), 0);

    reply_info.rcvid = ctp->rcvid;
    reply_info.offset = sizeof(msg->o);

    if (hci_queue_callback(iface, HCI_OGF_INFO, HCI_READ_BDADDR,
            0, reply_info, NULL, callback_get_bdaddr_complete) != EOK)
        return EAGAIN;

    hci_send_command(iface, HCI_OGF_INFO, HCI_READ_BDADDR, NULL, 0);

    return _RESMGR_NOREPLY;
}


static int iface_io_devctl(resmgr_context_t * ctp, io_devctl_t * msg,
        RESMGR_OCB_T * ocb)
{
    blue_iface_t * iface = (blue_iface_t *) ocb->hdr.attr;
    int status;

    if ((status = iofunc_devctl_default(ctp, msg,
                    (iofunc_ocb_t *) ocb)) != _RESMGR_DEFAULT) {
        return status;
    }

    switch (msg->i.dcmd) {
        case DCMD_BLUE_INQUIRY:
            return iface_io_inquiry(iface, ctp, msg);
        case DCMD_BLUE_REMOTE_NAME:
            return iface_io_remote_name(iface, ctp, msg);
        case DCMD_BLUE_LOCAL_NAME:
            return iface_io_local_name(iface, ctp, msg);
        case DCMD_BLUE_CHANGE_NAME:
            return iface_io_change_name(iface, ctp, msg);
        case DCMD_BLUE_GET_BDADDR:
            return iface_io_get_bdaddr(iface, ctp, msg);
        default:
            return ENOSYS;
    }

    return ENOSYS;
}

static int get_devnum()
{
    int            cnt;
    blue_iface_t  *iface;

    iofunc_attr_lock(&BlueCtrl.attr);
    for (cnt = 0; cnt < MAX_BLUE_DEVICES; cnt++) {
        for (iface = TAILQ_FIRST(&BlueCtrl.interfaces); iface; iface = TAILQ_NEXT(iface, dlink)) {
            if (iface->devnum == cnt)
                break;
        }
        if (!iface)
            break;
    }
    iofunc_attr_unlock(&BlueCtrl.attr);
    if (cnt == MAX_BLUE_DEVICES)
        return -1;

    return cnt;
}

static int iface_io_lock_ocb(resmgr_context_t * ctp, void * reserved,
        RESMGR_OCB_T * ocb)
{
    blue_iface_t * iface = (blue_iface_t *) ocb->hdr.attr;
    pthread_mutex_lock(&iface->mutex);
    return iofunc_attr_lock(ocb->hdr.attr);
}

static int iface_io_unlock_ocb(resmgr_context_t * ctp, void * reserved,
        RESMGR_OCB_T * ocb)
{
    blue_iface_t * iface = (blue_iface_t *) ocb->hdr.attr;
    pthread_mutex_unlock(&iface->mutex);
    return iofunc_attr_unlock(ocb->hdr.attr);
}

int iface_resmgr_attach(blue_iface_t * iface)
{
    uint8_t name[256];

    iface->tios.c_oflag = 0;
    iface->devnum = -1;
    iface->devnum = get_devnum();
    if (iface->devnum < 0)
        return EIO;

    iofunc_attr_init(&iface->attr, 0644 | S_IFCHR, 0, 0);
    iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &iface->resmgr_connect_funcs,
            _RESMGR_IO_NFUNCS, &iface->resmgr_io_funcs);

    if ((iface->attr.rdev = rsrcdbmgr_devno_attach("usb", -1, 0)) == -1) {
        SLOG(_SLOG_ERROR, "bluetooth: Unable to acquire resmgr devno");
        return EIO;
    }

    sprintf(name, "/dev/blue/hci%d", iface->devnum);
//    iface->resmgr_connect_funcs.open     = iface_io_open;
//    iface->resmgr_io_funcs.close_ocb     = iface_io_close_ocb;
    iface->resmgr_io_funcs.devctl        = iface_io_devctl;
    iface->resmgr_io_funcs.lock_ocb      = iface_io_lock_ocb;
    iface->resmgr_io_funcs.unlock_ocb    = iface_io_unlock_ocb;

    if ((iface->resmgr_id = resmgr_attach(BlueCtrl.dispatch, NULL, name, _FTYPE_ANY,
                    0, &iface->resmgr_connect_funcs, &iface->resmgr_io_funcs,
                    &iface->attr)) == -1) {
        SLOG(_SLOG_ERROR, "bluetooth: Failed to attach to resmgr");
        rsrcdbmgr_devno_detach(iface->attr.rdev, 0);
        return EIO;
    }

    resmgr_devino(iface->resmgr_id, &iface->io_mount.dev, &iface->attr.inode);
    return EOK;
}
    
void iface_resmgr_detach(blue_iface_t * iface)
{
    if (rsrcdbmgr_devno_detach(iface->attr.rdev, 0)) {
        SLOG(_SLOG_ERROR, "bluetooth: rsrcddbmgr_devno_detach failed");
    }
    if (resmgr_detach(BlueCtrl.dispatch, iface->resmgr_id,
                    _RESMGR_DETACH_ALL | _RESMGR_DETACH_CLOSE)) {
        SLOG(_SLOG_ERROR, "bluetooth: resmgr_detach failed");
    }
}
