/**
 * @file l2cap.c
 *
 * This is an implementation of the Logical Link Control and Adaptation
 * Protocol (L2CAP) which is intended to be run as a layer above the
 * HCI and Baseband Bluetooth protocols.  Its purpose is to provide
 * protocol multiplexing (as Baseband is limited to a single connection
 * per pair of hosts), segmentation/reassembly, and quality of service.
 *
 * This particular implementation uses standard POSIX sockets as the API
 * by which an application manages a Bluetooth connection.  The API is
 * designed to be compatable with that of the Bluez driver for Linux
 * (see http://bluez.sf.net).  In general, L2CAP allows for both
 * connectionless and connection-based traffic, although the Baseband
 * layer is solely connection-based.  So far, only connectionless
 * traffic is supported by this driver, using the SOCK_DGRAM socket type.
 * The SOCK_SEQPACKET type will be implemented to provide connection-
 * based sockets.
 *
 * Revision History:
 *   2003 Feb 4   David Moore      Initial version, providing the socket
 *                                 backend for SOCK_DGRAM sockets.
 */

#include <stdio.h>

/* These overrides must be defined before iofunc.h is included */
struct ocb_l2;
#define IOFUNC_OCB_T    struct ocb_l2

#include "blue_driver.h"
#include <bluetooth/bluetooth.h>
#include <sys/dcmd_ip.h>
#include <sys/sockmsg.h>
#include <fcntl.h>


static int l2cap_send_connless(hci_conn_t * conn, unsigned short psm,
        unsigned char * buf, unsigned int len)
{
    blue_iface_t * iface = conn->iface;
    l2cap_cl_pkt_t * pkt;
    int acl_len;

    pkt = (l2cap_cl_pkt_t *) malloc(iface->max_acl_data_len);
    if (!pkt) {
        return ENOMEM;
    }

    pkt->len  = ENDIAN_LE16(len + 2);
    pkt->chid = ENDIAN_LE16(0x0002);
    pkt->psm  = ENDIAN_LE16(psm);

    if (len > (iface->max_acl_data_len - 6))
        acl_len = iface->max_acl_data_len;
    else
        acl_len = len + 6;

    memcpy(pkt->data, buf, acl_len - 6);
    hci_send_acl(conn, 0, (unsigned char *) pkt, acl_len);
    len -= acl_len - 6;
    buf += acl_len - 6;

    while (len > 0) {
        acl_len = min(iface->max_acl_data_len, len);
        hci_send_acl(conn, HCI_ACL_CONT, buf, acl_len);

        len -= acl_len;
        buf += acl_len;
    }

    return EOK;
}

static int l2cap_recv_conless(hci_conn_t * conn, unsigned short psm,
        unsigned char * buf, unsigned int len)
{
    blue_t * blue = conn->iface->blue;
    IOFUNC_OCB_T * sock;

    iofunc_attr_lock(&blue->attr);

    sock = TAILQ_FIRST(&blue->sockets);
    while (sock) {
        if (sock->is_bound && sock->local_psm == psm
                && !bacmp(&sock->local_bdaddr, &conn->iface->bdaddr))
            break;

        sock = TAILQ_NEXT(sock, dlink);
    }

    if (!sock) {
        sock = TAILQ_FIRST(&blue->sockets);
        while (sock) {
            if (sock->is_bound && sock->local_psm == psm
                    && !bacmp(&sock->local_bdaddr, BDADDR_ANY))
                break;

            sock = TAILQ_NEXT(sock, dlink);
        }
    }

    if (sock) {
        if (sock->saved_rcvid != -1) {
            iov_t               iov[3];
            int                 reply_len = min(sock->saved_nbytes, len);
            io_sock_recvfrom_t  msg;
            struct sockaddr_l2  addr;

            msg.o.addrlen  = sock->saved_addrlen;
            addr.l2_family = AF_BLUETOOTH;
            addr.l2_bdaddr = conn->bdaddr;
            addr.l2_psm    = ENDIAN_LE16(psm);

            SETIOV(&iov[0], &msg.o, sizeof(msg.o));
            SETIOV(&iov[1], &addr, sock->saved_addrlen);
            SETIOV(&iov[2], buf, reply_len);

            MsgReplyv(sock->saved_rcvid, reply_len, iov, 3);
            sock->saved_rcvid = -1;
        }
        else {
            l2cap_pkt_t * pkt;

            pkt = (l2cap_pkt_t *) malloc(sizeof(l2cap_pkt_t) + len);
            if (!pkt) {
                iofunc_attr_unlock(&blue->attr);
                return ENOMEM;
            }

            pkt->len = len;
            pkt->addr.l2_family = AF_BLUETOOTH;
            pkt->addr.l2_bdaddr = conn->bdaddr;
            pkt->addr.l2_psm    = ENDIAN_LE16(psm);
            memcpy(pkt->data, buf, len);

            TAILQ_INSERT_TAIL(&sock->recv_pkts, pkt, dlink);

            iofunc_notify_trigger(blue->notify, 1, IOFUNC_NOTIFY_INPUT);
        }
    }
    else {
        SLOG(_SLOG_DEBUG1, "bluetooth: Ignoring packet with psm 0x%04x", psm);
    }

    iofunc_attr_unlock(&blue->attr);

    return EOK;
}

int l2cap_recv_packet(hci_conn_t * conn, unsigned char * buf,
        unsigned int len)
{
    l2cap_cl_pkt_t * pkt = (l2cap_cl_pkt_t *) buf;

    if (ENDIAN_LE16(pkt->chid) == 0x0002) {
        l2cap_recv_conless(conn, ENDIAN_LE16(pkt->psm), pkt->data,
                ENDIAN_LE16(pkt->len) - 2);
        return EOK;
    }

    SLOG(_SLOG_INFO, "bluetooth: Dropped connection-based packet");
    return EOK;
}


static blue_iface_t * l2cap_find_iface(blue_t * blue, RESMGR_OCB_T * ocb)
{
    blue_iface_t * iface;

    if (ocb->is_bound) {
        iface = hci_get_iface_by_bdaddr(blue, ocb->local_bdaddr);
    }
    else {
        iface = hci_get_default_iface(blue);
    }
    
    return iface;
}



/***************************************************************
 * The socket API                                              *
 ***************************************************************/

 
static int l2cap_io_open(resmgr_context_t * ctp, io_open_t * msg,
        RESMGR_HANDLE_T * handle, void * extra)
{
    io_socket_extra_t * param = (io_socket_extra_t *) extra;
    int status;

    status = iofunc_open_default(ctp, msg, handle, extra);

    if (status != EOK)
        return status;

    if (msg->connect.extra_type == _IO_CONNECT_EXTRA_SOCKET) {
        if (param->type != SOCK_DGRAM) {
            return ENOTSUP;
        }
        /* TODO: store packet type in the OCB somehow */

        if (param->protocol != BTPROTO_L2CAP) {
            return ENOTSUP;
        }
    }

    return EOK;
}

static int l2cap_io_close_ocb(resmgr_context_t * ctp, void * reserved,
        RESMGR_OCB_T * ocb)
{
    iofunc_notify_remove(ctp, ocb->blue->notify);

    return (iofunc_close_ocb_default(ctp, reserved, (iofunc_ocb_t *) ocb));
}

static int l2cap_io_connect(blue_t * blue, struct sockaddr_l2 * addr,
        RESMGR_OCB_T * ocb)
{
    hci_conn_t * conn;
    blue_iface_t * iface;

    if (ocb->state == SK_CONNECTED) {
        return EISCONN;
    }

    iface = l2cap_find_iface(blue, ocb);
    if (!iface) {
        return EHOSTUNREACH;
    }

    ocb->remote_bdaddr = addr->l2_bdaddr;
    ocb->remote_psm    = addr->l2_psm;
    ocb->state         = SK_CONNECTED;

    conn = hci_connect(iface, addr->l2_bdaddr);
    if (!conn) {
        return ENOMEM;
    }
    
    return EOK;
}

static int l2cap_io_bind(blue_t * blue, struct sockaddr_l2 * addr,
        RESMGR_OCB_T * ocb)
{
    RESMGR_OCB_T * sock;

    if (!hci_get_iface_by_bdaddr(blue, addr->l2_bdaddr)) {
        return ENODEV;
    }
    
    sock = TAILQ_FIRST(&blue->sockets);
    while (sock) {
        if (sock->is_bound && sock->local_psm == addr->l2_psm)
            return EADDRINUSE;
        sock = TAILQ_NEXT(sock, dlink);
    }

    ocb->local_bdaddr = addr->l2_bdaddr;
    ocb->local_psm    = addr->l2_psm;
    ocb->is_bound     = 1;

    return EOK;
}

static int l2cap_io_devctl(resmgr_context_t * ctp, io_devctl_t * msg,
        RESMGR_OCB_T * ocb)
{
    blue_t * blue = (blue_t *) ocb->blue;
    int status;
    struct sockaddr_l2 addr;

    if ((status = iofunc_devctl_default(ctp, msg,
                    (iofunc_ocb_t *) ocb)) != _RESMGR_DEFAULT) {
        return status;
    }

    switch (msg->i.dcmd) {
        case DCMD_IP_SDESTADDR:
            if (msg->i.nbytes != sizeof(struct sockaddr_l2))
                return EINVAL;
            
            resmgr_msgread(ctp, &addr, sizeof(struct sockaddr_l2), sizeof(msg->i));
            return l2cap_io_connect(blue, &addr, ocb);
        case DCMD_IP_SSRCADDR:
            if (msg->i.nbytes != sizeof(struct sockaddr_l2))
                return EINVAL;
            
            resmgr_msgread(ctp, &addr, sizeof(struct sockaddr_l2), sizeof(msg->i));
            return l2cap_io_bind(blue, &addr, ocb);
        default:
            return ENOSYS;
    }

    return ENOSYS;
}

static int l2cap_io_write(resmgr_context_t * ctp, io_write_t * msg,
        RESMGR_OCB_T * ocb)
{
    io_sock_sendto_t   *sendto = (io_sock_sendto_t *) msg;
    blue_iface_t       *iface;
    unsigned char      *buf;
    iov_t               iov[2];
    struct sockaddr_l2  addr;
    bdaddr_t            dest;
    unsigned short      psm;
    hci_conn_t         *conn;

    if ((msg->i.xtype & _IO_XTYPE_MASK) != _IO_XTYPE_TCPIP) {
        return ENOSYS;
    }

    iface = l2cap_find_iface(ocb->blue, ocb);
    if (!iface) {
        return EHOSTUNREACH;
    }

    if (ocb->pkt_type == SOCK_DGRAM
            && sendto->i.write.nbytes > L2CAP_CONLESS_MTU) {
        return EMSGSIZE;
    }
    /* TODO: this MTU will need to be negotiated */
    else if (sendto->i.write.nbytes > L2CAP_MTU) {
        return EMSGSIZE;
    }

    buf = (unsigned char *) malloc(sendto->i.write.nbytes);
    if (!buf) {
        return ENOMEM;
    }

    SETIOV(&iov[0], &addr, sendto->i.addrlen);
    SETIOV(&iov[1], buf, sendto->i.write.nbytes);
    resmgr_msgreadv(ctp, iov, 2, sendto->i.write.combine_len);

    if (sendto->i.addrlen == 0 && ocb->state == SK_CONNECTED) {
        dest = ocb->remote_bdaddr;
        psm  = ocb->remote_psm;
    }
    else if (sendto->i.addrlen == sizeof(struct sockaddr_l2)) {
        dest = addr.l2_bdaddr;
        psm  = addr.l2_psm;
    }
    else {
        free(buf);
        return ENOTCONN;
    }

    conn = hci_connect(iface, dest);
    if (!conn) {
        free(buf);
        return ENOMEM;
    }

    l2cap_send_connless(conn, psm, buf, sendto->i.write.nbytes);
    _IO_SET_WRITE_NBYTES(ctp, sendto->i.write.nbytes);

    free(buf);
    
    return EOK;
}

static int l2cap_io_read(resmgr_context_t * ctp, io_read_t * msg,
        RESMGR_OCB_T * ocb)
{
    io_sock_recvfrom_t *recvfrom = (io_sock_recvfrom_t *) msg;
    iov_t               iov[3];
    l2cap_pkt_t        *pkt;

    if ((msg->i.xtype & _IO_XTYPE_MASK) != _IO_XTYPE_TCPIP) {
        return ENOSYS;
    }

    if (recvfrom->i.addrlen != 0
            && recvfrom->i.addrlen != sizeof(struct sockaddr_l2)) {
        return EINVAL;
    }

    /* TODO: handle flags */

    if ((pkt = TAILQ_FIRST(&ocb->recv_pkts))) {
        io_sock_recvfrom_t rmsg;
        int nbytes = min(recvfrom->i.read.nbytes, pkt->len);

        TAILQ_REMOVE(&ocb->recv_pkts, pkt, dlink);

        rmsg.o.addrlen = recvfrom->i.addrlen;

        _IO_SET_READ_NBYTES(ctp, nbytes);

        SETIOV(&iov[0], &rmsg.o, sizeof(rmsg.o));
        SETIOV(&iov[1], &pkt->addr, recvfrom->i.addrlen);
        SETIOV(&iov[2], &pkt->data, nbytes);

        resmgr_msgwritev(ctp, iov, 3, 0);

        free(pkt);
        return EOK;
    }

    if (ocb->hdr.ioflag & O_NONBLOCK || ocb->saved_rcvid != -1) {
        return EAGAIN;
    }

    ocb->saved_rcvid = ctp->rcvid;
    ocb->saved_addrlen = recvfrom->i.addrlen;
    ocb->saved_nbytes = recvfrom->i.read.nbytes;
    
    return _RESMGR_NOREPLY;
}

static int l2cap_io_notify(resmgr_context_t * ctp, io_notify_t * msg,
        RESMGR_OCB_T * ocb)
{
    int trig;

    trig = _NOTIFY_COND_OUTPUT;
    if (TAILQ_FIRST(&ocb->recv_pkts))
        trig |= _NOTIFY_COND_INPUT;

    return (iofunc_notify(ctp, msg, ocb->blue->notify, trig, NULL, NULL));
}

static int l2cap_io_unblock(resmgr_context_t * ctp, io_pulse_t * msg,
        RESMGR_OCB_T * ocb)
{
    int status;

    status = iofunc_unblock_default(ctp, msg, (iofunc_ocb_t *) ocb);
    if (status != _RESMGR_DEFAULT) {
        return status;
    }

    if (ocb->saved_rcvid != -1) {
        MsgError(ocb->saved_rcvid, EINTR);
        ocb->saved_rcvid = -1;
        return _RESMGR_NOREPLY;
    }

    return _RESMGR_DEFAULT;
}


/***************************************************************
 * The Resource Manager                                        *
 ***************************************************************/


static IOFUNC_OCB_T * ocb_calloc(resmgr_context_t * ctp, IOFUNC_ATTR_T *device)
{
    struct ocb_l2 * ocb;

    if (!(ocb = calloc(1,sizeof(struct ocb_l2)))) {
        return NULL;
    }

    ocb->state = SK_UNCONNECTED;
    ocb->pkt_type = SOCK_DGRAM;
    ocb->blue = (blue_t *) device;
    ocb->remote_psm = 0;
    ocb->local_psm = 0;
    ocb->is_bound = 0;
    ocb->saved_rcvid = -1;
    TAILQ_INIT(&ocb->recv_pkts);

    TAILQ_INSERT_HEAD(&ocb->blue->sockets, ocb, dlink);
    
    return ocb;
}

static void ocb_free (IOFUNC_OCB_T * ocb)
{
    l2cap_pkt_t * pkt;

    while ((pkt = TAILQ_FIRST(&ocb->recv_pkts))) {
        TAILQ_REMOVE(&ocb->recv_pkts, pkt, dlink);
        free(pkt);
    }
    TAILQ_REMOVE(&ocb->blue->sockets, ocb, dlink);
    free(ocb);
}

static iofunc_funcs_t ocb_funcs = {
    _IOFUNC_NFUNCS,
    ocb_calloc,
    ocb_free
};

static iofunc_mount_t  mountpoint = {0, 0, 0, 0, &ocb_funcs };

int l2cap_resmgr_attach(blue_t * blue)
{
    blue->tios.c_oflag = 0;
    iofunc_attr_init(&blue->attr, 0666 | S_IFSOCK, 0, 0);
    TAILQ_INIT(&blue->sockets);
    IOFUNC_NOTIFY_INIT(blue->notify);
    iofunc_func_init( _RESMGR_CONNECT_NFUNCS, &blue->resmgr_connect_funcs,
            _RESMGR_IO_NFUNCS, &blue->resmgr_io_funcs);

    blue->attr.mount = &mountpoint;

    blue->resmgr_connect_funcs.open     = l2cap_io_open;
    blue->resmgr_io_funcs.close_ocb     = l2cap_io_close_ocb;
    blue->resmgr_io_funcs.devctl        = l2cap_io_devctl;
    blue->resmgr_io_funcs.write         = l2cap_io_write;
    blue->resmgr_io_funcs.read          = l2cap_io_read;
    blue->resmgr_io_funcs.notify        = l2cap_io_notify;
    blue->resmgr_io_funcs.unblock       = l2cap_io_unblock;

    if ((blue->resmgr_id = resmgr_attach(BlueCtrl.dispatch, NULL, "/dev/socket/31", _FTYPE_SOCKET,
                    0, &blue->resmgr_connect_funcs, &blue->resmgr_io_funcs,
                    &blue->attr)) == -1) {
        SLOG(_SLOG_ERROR, "bluetooth: Failed to attach to resmgr");
        return EIO;
    }

    resmgr_devino(blue->resmgr_id, &blue->io_mount.dev, &blue->attr.inode);
    return EOK;
}
    
void l2cap_resmgr_detach(blue_t * blue)
{
    if (resmgr_detach(BlueCtrl.dispatch, blue->resmgr_id,
                    _RESMGR_DETACH_ALL | _RESMGR_DETACH_CLOSE)) {
        SLOG(_SLOG_ERROR, "bluetooth: resmgr_detach failed");
    }
}
