/**
 * @file usb_io.c
 *
 * I/O primitives for USB transfers.
 *
 * Revision History:
 *   2003 Feb 4   David Moore      Initial version.
 */

#include <stdio.h>
#include "blue_driver.h"

int blue_cmd_io(blue_iface_t * iface, struct usbd_pipe * pipe, _uint32 flags,
        _uint8 * buffer, _uint32 length)
{
    _uint32 status;

    usbd_setup_vendor(iface->urb_cmd, flags, 0, USB_TYPE_CLASS,
            0, 0, buffer, length);
    status = usbd_io(iface->urb_cmd, pipe, NULL, iface, USBD_TIME_INFINITY);

    if (status != EOK)
        return status;

    return EOK;
}

static void blue_bulk_sync_callback(struct usbd_urb * urb, struct usbd_pipe * pipe,
        void * hdl)
{
    pthread_sleepon_lock();
    pthread_sleepon_signal(urb);
    pthread_sleepon_unlock();

    return;
}

int blue_bulk_io(blue_iface_t * iface, struct usbd_pipe * pipe, _uint32 flags,
        _uint8 * buffer, _uint32 length)
{
    _uint32 status, alen;

    usbd_setup_bulk(iface->urb_bulk_out, flags, buffer, length);
    status = usbd_io(iface->urb_bulk_out, pipe, blue_bulk_sync_callback,
            iface, USBD_TIME_INFINITY);

//    printf("io status = %d\n", status);
    if (status != EOK)
        return status;

    pthread_sleepon_lock();
    while (usbd_urb_status(iface->urb_bulk_out, &status, &alen) == EBUSY &&
            !(iface->status & BLUE_UNBLOCK)) {
        pthread_sleepon_wait(iface->urb_bulk_out);
    }
    pthread_sleepon_unlock();

    if ((status & USBD_USB_STATUS_MASK) == USBD_STATUS_STALL) {
        usbd_reset_pipe(pipe);
    }

    if (iface->status & BLUE_UNBLOCK) {
//        printf("INTR!\n");
        return EINTR;
    }

    if ((status & USBD_URB_STATUS_MASK) != USBD_STATUS_CMP) {
//        printf("IO! %x %x %x\n", status, USBD_USB_STATUS_MASK, USBD_STATUS_CMP);
        return EIO;
    }

    return EOK;
}

static void blue_intr_sync_callback(struct usbd_urb * urb, struct usbd_pipe * pipe,
        void * hdl)
{
//    printf("callback!\n");
    pthread_sleepon_lock();
    pthread_sleepon_signal(urb);
    pthread_sleepon_unlock();

    return;
}

int blue_intr_in_sync(blue_iface_t * iface)
{
    _uint32 status, alen;

    usbd_setup_interrupt(iface->urb_int, URB_DIR_IN | URB_SHORT_XFER_OK,
            iface->ep_int_buf, 16);
    status = usbd_io(iface->urb_int, iface->ep_int, blue_intr_sync_callback,
            iface, USBD_TIME_INFINITY);

    if (status != EOK) {
        return status;
    }

    pthread_sleepon_lock();
    while (usbd_urb_status(iface->urb_int, &status, &alen) == EBUSY &&
            !(iface->status & BLUE_UNBLOCK)) {
        pthread_sleepon_wait(iface->urb_int);
    }
    pthread_sleepon_unlock();
//    printf("intrsync status = %x, alen = %d\n", status, alen);

    if ((status & USBD_USB_STATUS_MASK) == USBD_STATUS_STALL) {
//        printf("reset!\n");
        usbd_reset_pipe(iface->ep_int);
    }

    if (iface->status & BLUE_UNBLOCK) {
//        printf("INTR!\n");
        return EINTR;
    }

    if ((status & USBD_URB_STATUS_MASK) != USBD_STATUS_CMP) {
//        printf("IO! %x %x %x\n", status, USBD_USB_STATUS_MASK, USBD_STATUS_CMP);
        return EIO;
    }

    return EOK;
}

static void blue_bulk_callback(struct usbd_urb * urb, struct usbd_pipe * pipe,
        void * hdl)
{
    blue_iface_t * iface = hdl;
    _uint32  ustatus, alen;
    int status;

//    printf("blue bulk callback!?!?!\n");
    if (!(iface->status & BLUE_PRESENT)) {
        return;
    }
    if ((status = usbd_urb_status(urb, &ustatus, &alen))) {
        SLOG(_SLOG_ERROR, "bluetooth: Bulk I/O callback failed");
        return;
    }
//    printf("bulk in %x %x %x\n", ustatus, alen, status);

    if ((ustatus & USBD_URB_STATUS_MASK) != USBD_STATUS_CMP) {
        if ((ustatus & USBD_USB_STATUS_MASK) == USBD_STATUS_STALL) {
            usbd_reset_pipe(pipe);
        }
        SLOG(_SLOG_ERROR, "bluetooth: Bulk in error, 0x%x", ustatus);
        iface->acl_buf_pos = 0;
    }
    else {
        unsigned short acl_len;

        memcpy(iface->acl_buf + iface->acl_buf_pos, iface->ep_bulk_buf, alen);
        iface->acl_buf_pos += alen;

        acl_len = ENDIAN_LE16(*(unsigned short *)&iface->acl_buf[2]);

        if (iface->acl_buf_pos - 4 == acl_len) {
            iface->hci_aclrecv(iface, iface->acl_buf, iface->acl_buf_pos);
            iface->acl_buf_pos = 0;
        }
        else if (iface->acl_buf_pos - 4 > acl_len) {
            SLOG(_SLOG_NOTICE, "bluetooth: Dropped large ACL packet");
            iface->acl_buf_pos = 0;
        }
    }

    if (iface->status & BLUE_RECV_ABORTED) {
        return;
    }

    if (iface->status & BLUE_PRESENT) {
        usbd_setup_bulk(urb, URB_DIR_IN | URB_SHORT_XFER_OK,
                iface->ep_bulk_buf, iface->ep_bulk_size);
        ustatus = usbd_io(urb, iface->ep_bulk_in, blue_bulk_callback, iface,
                USBD_TIME_INFINITY);
        if (ustatus) {
            SLOG(_SLOG_ERROR, "bluetooth: Failed to restart bulk callback");
        }
    }
}

static void blue_intr_callback(struct usbd_urb * urb, struct usbd_pipe * pipe,
        void * hdl)
{
    blue_iface_t * iface = hdl;
    _uint32  ustatus, alen;
    int status;

    if (!(iface->status & BLUE_PRESENT)) {
        return;
    }
//    printf("int\n");
    if ((status = usbd_urb_status(urb, &ustatus, &alen))) {
        SLOG(_SLOG_ERROR, "bluetooth: Interrupt I/O callback failed");
        return;
    }

    if ((ustatus & USBD_URB_STATUS_MASK) != USBD_STATUS_CMP) {
        if ((ustatus & USBD_USB_STATUS_MASK) == USBD_STATUS_STALL) {
            usbd_reset_pipe(pipe);
        }
        SLOG(_SLOG_ERROR, "bluetooth: Interrupt in error, 0x%x", ustatus);
        iface->evt_buf_pos = 0;
    }
    else {
        memcpy(iface->evt_buf + iface->evt_buf_pos, iface->ep_int_buf, alen);
        iface->evt_buf_pos += alen;

        if (iface->evt_buf_pos - 2 == iface->evt_buf[1]) {
            iface->hci_event(iface, iface->evt_buf, iface->evt_buf_pos);
            iface->evt_buf_pos = 0;
        }
        else if (iface->evt_buf_pos - 2 > iface->evt_buf[1]) {
            SLOG(_SLOG_NOTICE, "bluetooth: Dropped large event packet");
            iface->evt_buf_pos = 0;
        }
    }

    if (iface->status & BLUE_PRESENT) {
        usbd_setup_interrupt(urb, URB_DIR_IN | URB_SHORT_XFER_OK,
                iface->ep_int_buf, iface->ep_int_size);
        ustatus = usbd_io(urb, iface->ep_int, blue_intr_callback, iface,
                USBD_TIME_INFINITY);
        if (ustatus) {
            SLOG(_SLOG_ERROR, "bluetooth: Failed to restart interrupt callback");
        }
    }
}

int blue_start_ep_listeners(blue_iface_t * iface)
{
    int status;

    iface->evt_buf_pos = 0;
    iface->evt_buf = (_uint8 *) malloc(MAX_RECV_EVT_LEN);
    if (!iface->evt_buf)
        return ENOMEM;

    iface->acl_buf_pos = 0;
    iface->acl_buf = (_uint8 *) malloc(MAX_RECV_ACL_LEN);
    if (!iface->acl_buf)
        return ENOMEM;

    usbd_setup_interrupt(iface->urb_int, URB_DIR_IN | URB_SHORT_XFER_OK,
            iface->ep_int_buf, iface->ep_int_size);
    status = usbd_io(iface->urb_int, iface->ep_int, blue_intr_callback, iface,
            USBD_TIME_INFINITY);

    if (status != EOK) {
        return status;
    }

    blue_start_acl_listen(iface);

    return status;
}

int blue_start_acl_listen(blue_iface_t * iface)
{
    int status;

    iface->status &= ~BLUE_RECV_ABORTED;

    usbd_setup_bulk(iface->urb_bulk_in, URB_DIR_IN | URB_SHORT_XFER_OK,
            iface->ep_bulk_buf, iface->ep_bulk_size);
    status = usbd_io(iface->urb_bulk_in, iface->ep_bulk_in, blue_bulk_callback, iface,
            USBD_TIME_INFINITY);

    return status;
}

int blue_abort_acl_listen(blue_iface_t * iface)
{
    iface->status |= BLUE_RECV_ABORTED;
    usbd_abort_pipe(iface->ep_bulk_in);
    return EOK;
}
