#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

#include <bluetooth/bluetooth.h>

#define SET_NAME    1
#define INQUIRY     2

#define USAGE \
"Usage:\n" \
"    hciconfig interface [parameters]\n" \
"\n" \
"Parameters:\n" \
"    name <string>              Set local Bluetooth device name\n" \
"    inquiry                    Print the inquiry cache of nearby devices\n"


static void do_inquiry(int fd, int max_resp)
{
    int i, size;
    dcmd_blue_inquiry_t * dcmd;

    size = sizeof(dcmd_blue_inquiry_t) + max_resp * sizeof(inq_res_t);
    dcmd = (dcmd_blue_inquiry_t *) malloc(size);
    if (!dcmd) {
        fprintf(stderr, "Out of memory\n");
        return;
    }

    dcmd->num_responses = max_resp; 
    if (devctl(fd, DCMD_BLUE_INQUIRY, dcmd, size, NULL) != EOK) {
        fprintf(stderr, "devctl(DCMD_BLUE_INQUIRY) failed: %s\n", strerror(errno));
        return;
    }

    printf("Cached %d inquiry response(s):\n", dcmd->num_responses);
    for (i=0; i<dcmd->num_responses; i++) {
        char str[32];
        ba2str(&dcmd->result[i].bdaddr, str);
        printf("  BDaddr %s  Class: %06x  Clk Offset: %05d\n",
                str, dcmd->result[i].class_of_device, dcmd->result[i].clock_offset);
    }

    free(dcmd);
}

struct param_info {
    int     code;
    char *  name;
};

static struct param_info params[] = {
    { SET_NAME,     "name" },
    { INQUIRY,      "inquiry" },
    { 0 }
};

int get_param(char * name)
{
    int i;

    for (i=0; params[i].code; i++) {
        if (!strcmp(name, params[i].name))
            return params[i].code;
    }

    return 0;
}

int main(int argc, char ** argv)
{
    int fd, i;
    int inquiry = 0;
    int quiet = 0;
    char * iface;
    char * set_name = NULL;
    char filename[256];

    if (argc < 2) {
        fprintf(stderr, "Error: Interface name must be specified.\n");
        fprintf(stderr, "%s", USAGE);
        return 1;
    }

    iface = argv[1];

    for (i=2; i < argc; i++) {
        switch (get_param(argv[i])) {
            case SET_NAME:
                if (++i >= argc) {
                    fprintf(stderr, "Error: \"name\" expects argument\n");
                    return 1;
                }
                set_name = argv[i];
                break;
            case INQUIRY:
                inquiry = 1;
                break;
            default:
                fprintf(stderr, "Error: Invalid parameter \"%s\".", argv[i]);
                return 1;
        }
    }

    snprintf(filename, sizeof(filename), "/dev/blue/%s", iface);
    fd = open(filename, O_RDONLY);
    if (fd < 0) {
        fprintf(stderr, "Error: Could not open interface \"%s\": %s\n",
                iface, strerror(errno));
        return 1;
    }

    if (set_name) {
        if (devctl(fd, DCMD_BLUE_CHANGE_NAME, set_name, strlen(set_name), NULL) != EOK) {
            fprintf(stderr, "devctl(DCMD_BLUE_CHANGE_NAME) failed: %s\n", strerror(errno));
        }
        quiet = 1;
    }

    if (inquiry) {
        do_inquiry(fd, 20);
        quiet = 1;
    }


    if (!quiet) {
        char bdaddr_str[32];
        bdaddr_t bdaddr;
        char name[256];

        if (devctl(fd, DCMD_BLUE_GET_BDADDR, &bdaddr, sizeof(bdaddr), NULL) != EOK) {
            fprintf(stderr, "devctl(DCMD_BLUE_GET_BDADDR) failed: %s\n", strerror(errno));
            close(fd);
            return 1;
        }
        ba2str(&bdaddr, bdaddr_str);

        if (devctl(fd, DCMD_BLUE_LOCAL_NAME, name, sizeof(name), NULL) != EOK) {
            fprintf(stderr, "devctl(DCMD_BLUE_LOCAL_NAME) failed: %s\n", strerror(errno));
            close(fd);
            return 1;
        }

        printf("%s:   UP  BDaddr %s  name: \"%s\"\n", iface, bdaddr_str, name);
    }

    close(fd);

    return 0;
}
