#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/l2cap.h>

#define TEST_RECEIVE        1
#define TEST_SEND           2
#define TEST_INQUIRY        3
#define TEST_REMOTE_NAME    4
#define TEST_LOCAL_NAME     5

#define USAGE \
"Usage:\n" \
"   test_blue -r [-p psm]                      wait for incoming packets\n" \
"   test_blue -s bdaddr [-p psm] string        send a string packet\n" \
"   test_blue -i                               print latest inquiry results\n" \
"   test_blue -n bdaddr                        get name of remote device\n" \
"   test_blue -l [name]                        get/set name of local device\n" \
"\n\n" \
"<bdaddr> is a colon seperated 6-tuple which is the bluetooth device\n" \
"    address of the remote receiver.\n" \
"<psm> is a Protocol/Service Multiplexer that must be odd and the 9th bit\n" \
"    must be 0.  (default is 0x1001)\n"


static void test_inquiry()
{
    int fd;
    int i, size;
    char * file = "/dev/blue/hci0";
    dcmd_blue_inquiry_t * dcmd;

    fd = open(file, O_RDONLY);
    if (fd < 0) {
        fprintf(stderr, "Could not open device %s\n", file);
        return;
    }

    size = sizeof(dcmd_blue_inquiry_t) + 10 * sizeof(inq_res_t);
    dcmd = (dcmd_blue_inquiry_t *) malloc(size);
    if (!dcmd) {
        fprintf(stderr, "Out of memory\n");
        close(fd);
        return;
    }

    dcmd->num_responses = 10;
    if (devctl(fd, DCMD_BLUE_INQUIRY, dcmd, size, NULL) != EOK) {
        fprintf(stderr, "devctl(DCMD_BLUE_INQUIRY) failed\n");
        close(fd);
        return;
    }

    printf("Received %d inquiry responses:\n", dcmd->num_responses);
    for (i=0; i<dcmd->num_responses; i++) {
        char str[32];
        ba2str(&dcmd->result[i].bdaddr, str);
        printf("  %02d.  %s\n", i+1, str);
    }

    free(dcmd);
    close(fd);
}

static void test_remote_name(bdaddr_t bdaddr)
{
    int fd;
    char * file = "/dev/blue/hci0";
    dcmd_blue_remote_name_t dcmd;

    fd = open(file, O_RDONLY);
    if (fd < 0) {
        fprintf(stderr, "Could not open device %s\n", file);
        return;
    }

    dcmd.bdaddr = bdaddr;
    if (devctl(fd, DCMD_BLUE_REMOTE_NAME, &dcmd, sizeof(dcmd), NULL) != EOK) {
        fprintf(stderr, "devctl(DCMD_BLUE_REMOTE_NAME) failed\n");
        close(fd);
        return;
    }

    printf("Remote name is \"%s\"\n", dcmd.remote_name);
    close(fd);
}

static void test_local_name(char * name)
{
    int fd;
    char * file = "/dev/blue/hci0";
    char str[32];

    fd = open(file, O_RDONLY);
    if (fd < 0) {
        fprintf(stderr, "Could not open device %s\n", file);
        return;
    }

    if (name) {
        if (devctl(fd, DCMD_BLUE_CHANGE_NAME, name, strlen(name), NULL) != EOK) {
            fprintf(stderr, "devctl(DCMD_BLUE_CHANGE_NAME) failed\n");
            close(fd);
            return;
        }

        printf("Changed local name to \"%s\"\n", name);
    }
    else {
        if (devctl(fd, DCMD_BLUE_LOCAL_NAME, str, sizeof(str), NULL) != EOK) {
            fprintf(stderr, "devctl(DCMD_BLUE_LOCAL_NAME) failed\n");
            close(fd);
            return;
        }

        printf("Local name is \"%s\"\n", str);
    }
    close(fd);
}

int main(int argc, char ** argv)
{
    struct sockaddr_l2 addr;
    int mode=0, fd;
    int ret;
    unsigned int psm = 0x1001;
    char opt;
    char * str;
    bdaddr_t bdaddr;

    while ((opt = getopt(argc, argv, "rils:p:n:")) != -1) {
        switch (opt) {
            case 'r':
                mode = TEST_RECEIVE;
                break;
            case 's':
                mode = TEST_SEND;
                str2ba(optarg, &bdaddr);
                break;
            case 'n':
                mode = TEST_REMOTE_NAME;
                str2ba(optarg, &bdaddr);
                break;
            case 'i':
                mode = TEST_INQUIRY;
                break;
            case 'l':
                mode = TEST_LOCAL_NAME;
                break;
            case 'p':
                psm = strtoul(optarg, NULL, 0);
                break;
            default:
                break;
        }
    }

    if (mode == 0) {
        fprintf(stderr, "%s", USAGE);
        return 1;
    }

    fd = socket(AF_BLUETOOTH, SOCK_DGRAM, BTPROTO_L2CAP);
    if (fd == -1) {
        fprintf(stderr, "Could not open socket: %s\n", strerror(errno));
        return 1;
    }

    switch (mode) {
        case TEST_RECEIVE:
            addr.l2_family = AF_BLUETOOTH;
            bacpy(&addr.l2_bdaddr, BDADDR_ANY);
            addr.l2_psm    = htobs(psm);

            if (bind(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
                fprintf(stderr, "Could not bind socket: %s\n",
                        strerror(errno));
                break;
            }

            while (1) {
                fd_set readfds;
                int size = sizeof(addr);
                char inbuf[1024];

                FD_ZERO(&readfds);
                FD_SET(fd, &readfds);
                if (select(fd+1, &readfds, NULL, NULL, NULL) < 0) {
                    printf("select() failed: %s\n", strerror(errno));
                    break;
                }
                
                ret = recvfrom(fd, inbuf, sizeof(inbuf), 0,
                        (struct sockaddr *) &addr, &size);
                if (ret >= 0) {
                    printf("Received packet: \"%.*s\"\n", ret, inbuf);
                }
                else {
                    printf("Failed to receive: %s\n", strerror(errno));
                    break;
                }
            }
            break;

        case TEST_SEND:
            addr.l2_family = AF_BLUETOOTH;
            addr.l2_bdaddr = bdaddr;
            addr.l2_psm    = htobs(psm);

            if (optind < argc)
                str = argv[optind];
            else
                str = "testing 1 2 3";

            ret = sendto(fd, str, strlen(str)+1, 0,
                    (struct sockaddr *) &addr, sizeof(addr));
            if (ret >= 0) {
                printf("Sent message of length %d\n", ret);
            }
            else {
                printf("Send failed: %s\n", strerror(errno));
            }
            break;
        case TEST_INQUIRY:
            test_inquiry();
            break;
        case TEST_REMOTE_NAME:
            test_remote_name(bdaddr);
            break;
        case TEST_LOCAL_NAME:
            if (optind < argc)
                test_local_name(argv[optind]);
            else
                test_local_name(NULL);
            break;
    }

    close(fd);

    return 0;
}
